/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.interceptor.AroundInvoke;
import javax.interceptor.InvocationContext;

import org.springframework.util.ObjectUtils;
import org.springframework.util.ReflectionUtils;

/**
 * Utility class used for finding annotation using the inheritance rules from
 * jsr250.
 * 
 * @author Costin Leau
 * 
 */
public abstract class JeeAnnotationUtils {

	public static final Class[] AROUND_INVOKE_METHOD_PARAM = new Class[] { InvocationContext.class };

	/**
	 * Simple which searches for an annotation inside an array.
	 * 
	 * @param <A>
	 * @param annotations
	 * @param annotationClass
	 * @return
	 */
	private static <A extends Annotation> A searchAnnotation(Annotation[] annotations, Class<A> annotationClass) {
		for (Annotation annotation : annotations) {
			if (annotation.annotationType() == annotationClass)
				return (A) annotation;
		}
		return null;
	}

	/**
	 * Apply the jsr250 annotation algorithm to retrieve the annotation value
	 * for the given member. This method has to be called by after the lookup
	 * for the member has been executed.
	 * 
	 * @param <A>
	 * @param annotationClass
	 * @param enclosingClass
	 * @param member
	 * @return
	 */
	public static <A extends Annotation> A findInheritedMemberAnnotation(Class<A> annotationClass,
			Class enclosingClass, AccessibleObject member) {

		// get member annotation
		A annotation = member.getAnnotation(annotationClass);

		if (annotation == null) {
			// search target class annotation (make sure they are not inherited)
			A classAnnotation = searchAnnotation(enclosingClass.getDeclaredAnnotations(), annotationClass);

			if (classAnnotation != null) {

				// check if it was not excluded by any member

				// a. check fields
				for (Field field : enclosingClass.getDeclaredFields()) {
					// exclude annotation
					if (field.isAnnotationPresent(annotationClass))
						return null;
				}

				// b. check methods
				for (Method method : enclosingClass.getDeclaredMethods()) {
					// exclude annotation
					if (method.isAnnotationPresent(annotationClass))
						return null;
				}
				annotation = classAnnotation;
			}
		}

		return annotation;
	}

	/**
	 * Returns the annotation for the given field (specified by name and type)
	 * using the inheritance mechanism defined in JSR 250. If the annotation is
	 * not found in the target class, the super classes will be searched for
	 * annotation using the jsr250 specs inheritance mechanism.
	 * 
	 * @param <A>
	 * @param annotationClass annotation class to look for
	 * @param targetClass target class to look at
	 * @param fieldName field name
	 * @param fieldType field type (optional - can be null)
	 * @return null if no annotation was found using the algorithm, not null
	 *         otherwise
	 */
	public static <A extends Annotation> A findFieldAnnotation(Class<A> annotationClass, Class targetClass,
			String fieldName, Class fieldType) {

		Field field = null;
		do {
			try {
				// search member
				Field localField = targetClass.getDeclaredField(fieldName);
				if (localField != null) {
					// check also field type
					if (fieldType != null && localField.getType() == fieldType)
						field = localField;
				}
			}
			catch (NoSuchFieldException e) {
				// move on
			}
			targetClass = targetClass.getSuperclass();
		} while (field == null && targetClass != null && targetClass != Object.class);

		A annotation = null;

		if (field != null)
			annotation = findInheritedMemberAnnotation(annotationClass, targetClass, field);
		return annotation;
	}

	/**
	 * Returns the annotation for the given method (specified by name and
	 * parameterTypes) using the inheritance mechanism defined in JSR 250. If
	 * the annotation is not found in the target class, the super classes will
	 * be searched for the annotation using the jsr250 spec inheritance
	 * mechanism.
	 * 
	 * @param <A>
	 * @param annotationClass
	 * @param targetClass
	 * @param methodName
	 * @param parameterTypes
	 * @return
	 */
	public static <A extends Annotation> A findMethodAnnotation(Class<A> annotationClass, Class targetClass,
			String methodName, Class... parameterTypes) {

		// allow null paramTypes arrays to be passed in.
		Class[] paramTypes = (parameterTypes == null ? new Class[] {} : parameterTypes);

		Method method = null;
		do {
			try {
				method = targetClass.getDeclaredMethod(methodName, paramTypes);
			}
			catch (NoSuchMethodException e) {
				// move on
				targetClass = targetClass.getSuperclass();
			}
		} while (method == null && targetClass != null && targetClass != Object.class);

		A annotation = null;
		if (method != null)
			annotation = findInheritedMemberAnnotation(annotationClass, targetClass, method);
		return annotation;

	}

	public static <A extends Annotation> A findMethodAnnotation(Class<A> annotationClass, Class targetClass,
			String methodName) {
		return findMethodAnnotation(annotationClass, targetClass, methodName, (Class[]) null);
	}

	/**
	 * Finds the annotation for a method. It does not use JSR-250 inheritance
	 * (ignores class level annotations and looks only at method level).
	 * 
	 * @param <A>
	 * @param annotationClass
	 * @param method
	 * @param clazz
	 * @return
	 */
	public static <A extends Annotation> A findMethodAnnotation(Class<A> annotationClass, Method method, Class clazz) {

		if (!annotationClass.isAnnotation()) {
			throw new IllegalArgumentException(annotationClass + " is not an annotation");
		}
		A annotation = method.getAnnotation(annotationClass);

		Class cl = clazz;
		while (annotation == null) {
			cl = cl.getSuperclass();
			if (cl == null || cl.equals(Object.class)) {
				break;
			}
			try {
				method = cl.getDeclaredMethod(method.getName(), (Class[]) method.getParameterTypes());
				annotation = method.getAnnotation(annotationClass);
			}
			catch (NoSuchMethodException ex) {
				// We're done
			}
		}
		return annotation;
	}

	/**
	 * Find all the methods on the given class and superclasses that have the
	 * given annotation.
	 * 
	 * @param clazz
	 * @param A
	 * @return
	 */
	public static List<Method> findAnnotatedMethods(final Class clazz, final Class<? extends Annotation> A) {
		final List<Method> found = new ArrayList<Method>();
		if (clazz == null)
			return found;
		ReflectionUtils.doWithMethods(clazz, new ReflectionUtils.MethodCallback() {
			public void doWith(Method m) throws IllegalArgumentException, IllegalAccessException {
				if (m.isAnnotationPresent(A)) {
					found.add(m);
				}
			}
		});
		return found;
	}

	/**
	 * Find all the methods on the given class and superclasses that have the
	 * given annotation and the given signature (jsr250 is not applied).
	 * 
	 * @param clazz
	 * @param A
	 * @param argTypes
	 * @return
	 */
	public static List<Method> findAnnotatedMethods(final Class clazz, final Class<? extends Annotation> A,
			final Class... argTypes) {
		final List<Method> found = new ArrayList<Method>();
		if (clazz == null)
			return found;
		ReflectionUtils.doWithMethods(clazz, new ReflectionUtils.MethodCallback() {
			public void doWith(Method m) throws IllegalArgumentException, IllegalAccessException {
				if (m.isAnnotationPresent(A) && ObjectUtils.nullSafeEquals(argTypes, m.getParameterTypes())) {
					found.add(m);
				}
			}
		});
		return found;
	}

	/**
	 * Find all methods with the given annotation and argument types on the
	 * given class and all its superclasses.
	 * 
	 * @param clazz component class to introspect
	 * @param A class of annotation to find
	 * @param argTypes types of arguments to expect
	 * @return a no-arg method with this annotation
	 */
	public static Method findFirstAnnotatedMethod(final Class clazz, final Class<? extends Annotation> A,
			final Class... argTypes) {

		Class targetClass = clazz;
		// Keep backing up the inheritance hierarchy.
		do {
			Method[] methods = targetClass.getDeclaredMethods();
			for (int i = 0; i < methods.length; i++) {
				Method m = methods[i];

				if (m.isAnnotationPresent(A) && ObjectUtils.nullSafeEquals(argTypes, m.getParameterTypes()))
					return m;

			}
			targetClass = targetClass.getSuperclass();
		} while (targetClass != null && targetClass != Object.class);

		return null;
	}

	/**
	 * Find the EJB3 around invoke method
	 */
	public static Method findAroundInvokeMethod(Class clazz) {
		return findFirstAnnotatedMethod(clazz, AroundInvoke.class, AROUND_INVOKE_METHOD_PARAM);
	}

}
