/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.config;

import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.BeanFactoryPostProcessor;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.jee.inject.BeanDefinitionAnnotationComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Enricher;
import org.springframework.jee.intercept.InterceptionMetadataBeanPostProcessor;

/**
 * Spring BeanFactoryBeanPostProcessor to automatically apply JEE injection and
 * interception to annotated classes deployed in a Spring context.
 * 
 * @author Rod Johnson
 */
public class JeeBeanFactoryPostProcessor implements BeanFactoryPostProcessor, ApplicationContextAware {

	private ApplicationContext owningContext;

	private DeploymentUnitMetadata deploymentUnitMetadata;

	private ServiceEnvironment serviceEnvironment;

	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
		this.owningContext = applicationContext;
	}

	public void postProcessBeanFactory(ConfigurableListableBeanFactory beanFactory) throws BeansException {
		if (!(beanFactory instanceof BeanDefinitionRegistry))
			throw new IllegalArgumentException("a beanFactory that implements "
					+ BeanDefinitionRegistry.class.getName() + " is required");

		BeanDefinitionRegistry bdr = (BeanDefinitionRegistry) beanFactory;

		this.deploymentUnitMetadata = new DeploymentUnitMetadata();
		deploymentUnitMetadata.setEnvironment(serviceEnvironment);

		Enricher enricher = new Enricher(owningContext, bdr);
		BeanDefinitionAnnotationComponentContributor bdacc = createComponentContributor(deploymentUnitMetadata);

		for (String beanDefinitionName : beanFactory.getBeanDefinitionNames()) {
			BeanDefinition bd = beanFactory.getBeanDefinition(beanDefinitionName);
			if (bd instanceof AbstractBeanDefinition) {
				bdacc.register(beanDefinitionName, (AbstractBeanDefinition) bd);
			}
		}

		bdacc.contribute(enricher);

		// Add the BeanPostProcessor that will process these definitions
		MetadataDrivenBeanPostProcessorSupport jeePostProcessor = createMetadataBeanPostProcessor();
		jeePostProcessor.setBeanFactory(beanFactory);
		beanFactory.addBeanPostProcessor(jeePostProcessor);
	}

	/**
	 * Hook for post processor creation.
	 * 
	 * @return
	 */
	protected MetadataDrivenBeanPostProcessorSupport createMetadataBeanPostProcessor() {
		return new InterceptionMetadataBeanPostProcessor();
	}

	/**
	 * Hook for bean-aware, component contributor.
	 * 
	 * @return
	 */
	protected BeanDefinitionAnnotationComponentContributor createComponentContributor(DeploymentUnitMetadata deploymentUnitMetadata) {
		return new BeanDefinitionAnnotationComponentContributor(deploymentUnitMetadata);
	}

	/**
	 * @return Returns the serviceEnvironment.
	 */
	public ServiceEnvironment getServiceEnvironment() {
		return serviceEnvironment;
	}

	/**
	 * @param serviceEnvironment The serviceEnvironment to set.
	 */
	public void setServiceEnvironment(ServiceEnvironment serviceEnvironment) {
		this.serviceEnvironment = serviceEnvironment;
	}

}
