/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.ejb;

import java.lang.reflect.Field;
import java.lang.reflect.Method;

import javax.ejb.EJB;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.jee.inject.MetadataProcessor;
import org.springframework.util.ReflectionUtils;

/**
 * Annotation processor for @EJB injection.
 * 
 * @author Costin Leau
 * 
 */
public class EjbReferenceProcessor implements MetadataProcessor {

	private static final Log log = LogFactory.getLog(EjbReferenceProcessor.class);

	public void process(Jsr250Metadata jsr250) {
		if (jsr250 instanceof EjbInjectionMetadata) {
			addEjbReferenceMetadata((EjbInjectionMetadata) jsr250);
		}
	}

	/**
	 * Process the current class and add ejb reference information.
	 * 
	 * @param EjbInjectionMetadata
	 */
	protected void addEjbReferenceMetadata(EjbInjectionMetadata EjbInjectionMetadata) {
		addClassLevelReferences(EjbInjectionMetadata);
		addMethodLevelReferences(EjbInjectionMetadata);
		addFieldLevelReferences(EjbInjectionMetadata);
	}

	/**
	 * Process class level EJB annotation
	 * 
	 * @param EjbInjectionMetadata
	 */
	private void addClassLevelReferences(EjbInjectionMetadata EjbInjectionMetadata) {
	}

	/**
	 * Process method level EJB annotation
	 * 
	 * @param EjbInjectionMetadata
	 */
	private void addMethodLevelReferences(final EjbInjectionMetadata EjbInjectionMetadata) {
		ReflectionUtils.doWithMethods(EjbInjectionMetadata.getComponentClass(), new ReflectionUtils.MethodCallback() {
			public void doWith(Method m) throws IllegalArgumentException, IllegalAccessException {
				processCandidateInjectionMethod(EjbInjectionMetadata, m);
			}
		});
	}

	private void processCandidateInjectionMethod(EjbInjectionMetadata EjbInjectionMetadata, Method m) {
		EJB reference = (EJB) JeeAnnotationUtils.findMethodAnnotation(EJB.class, m,
				EjbInjectionMetadata.getComponentClass());
		if (reference != null) {
			EjbMethodInjection mi = new EjbMethodInjection(m, reference);
			EjbInjectionMetadata.addEjbInjection(mi);
		}
	}

	/**
	 * Process field level EJB annotation
	 * 
	 * @param EjbInjectionMetadata
	 */
	private void addFieldLevelReferences(final EjbInjectionMetadata EjbInjectionMetadata) {
		ReflectionUtils.doWithFields(EjbInjectionMetadata.getComponentClass(), new ReflectionUtils.FieldCallback() {
			public void doWith(Field f) throws IllegalArgumentException, IllegalAccessException {
				processCandidateInjectionField(EjbInjectionMetadata, f);
			}
		});
	}

	private void processCandidateInjectionField(EjbInjectionMetadata EjbInjectionMetadata, Field f) {
		EJB reference = f.getAnnotation(EJB.class);
		if (reference != null) {
			EjbFieldInjection fi = new EjbFieldInjection(f, reference);
			EjbInjectionMetadata.addEjbInjection(fi);
		}
	}

}
