/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.inject;

import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.util.ClassUtils;

/**
 * Default ComponentContributor. Requires bean definitions to be registered
 * before contributing.
 * 
 * @author Rod Johnson
 * 
 */
public class BeanDefinitionAnnotationComponentContributor extends ChainedProcessorsComponentContributor {

	private static final Log log = LogFactory.getLog(BeanDefinitionAnnotationComponentContributor.class);

	DeploymentUnitMetadata deploymentUnitMetadata;

	private Map<String, AbstractBeanDefinition> beanDefinitions = new LinkedHashMap<String, AbstractBeanDefinition>();

	public BeanDefinitionAnnotationComponentContributor(DeploymentUnitMetadata deploymentUnitMetadata) {
		this.deploymentUnitMetadata = deploymentUnitMetadata;
	}

	public void register(String beanName, AbstractBeanDefinition bd) {
		this.beanDefinitions.put(beanName, bd);
	}

	public void registerMap(Map<String, AbstractBeanDefinition> definitionMap) {
		this.beanDefinitions.putAll(definitionMap);
	}

	public void contribute(Enricher enricher) {
		boolean debug = log.isDebugEnabled();

		for (String beanName : beanDefinitions.keySet()) {
			AbstractBeanDefinition definition = beanDefinitions.get(beanName);
			if (debug) {
				log.debug("Examining metadata for " + definition);
			}

			try {
				// The bean class will probably not yet have been resolved
				Class resolvedBeanClass = (definition.hasBeanClass() ? definition.getBeanClass() : ClassUtils
						.forName(definition.getBeanClassName()));
				enricher.attach(findComponentMetadata(beanName, resolvedBeanClass, deploymentUnitMetadata));
			}
			catch (IllegalStateException ex) {
				throw new IllegalStateException("Cannot resolve bean class for bean named '" + beanName + "'", ex);
			}
			catch (ClassNotFoundException ex) {
				throw new IllegalArgumentException("Cannot resolve bean class for bean named '" + beanName + "'", ex);
			}
		}
	}
}
