/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jee.transaction.TransactionalMetadata;

/**
 * Support for annotation-based component contributors that uses internally a
 * chain of metadata processors.
 * 
 * @MetadataProcessor
 * 
 * @author Rod Johnson
 * @author Costin Leau
 */
public abstract class ChainedProcessorsComponentContributor implements ComponentContributor {

	protected final Log log = LogFactory.getLog(getClass());

	/**
	 * Default list of metadata processors. Contains the Jsr250, Interception
	 * and Transaction processor.
	 */
	public static final List<MetadataProcessor> DEFAULT_PROCESSORS = new ArrayList<MetadataProcessor>(3);

	/**
	 * Jsr250 only processing chain.
	 */
	public static final List<MetadataProcessor> JSR250_PROCESSORS = new ArrayList<MetadataProcessor>(1);

	/**
	 * Jsr250 and Interception chain.
	 */
	public static final List<MetadataProcessor> INTERCEPTION_PROCESSORS = new ArrayList<MetadataProcessor>(2);

	static {
		// jsr250
		Jsr250Processor jsr250 = new Jsr250Processor();
		// ejb3 interceptors
		MetadataProcessor interception = new InterceptionProcessor(jsr250);
		// ejb3 transactions
		TransactionProcessor transaction = new TransactionProcessor();

		DEFAULT_PROCESSORS.add(jsr250);
		DEFAULT_PROCESSORS.add(interception);
		DEFAULT_PROCESSORS.add(transaction);

		JSR250_PROCESSORS.add(jsr250);

		INTERCEPTION_PROCESSORS.add(jsr250);
		INTERCEPTION_PROCESSORS.add(interception);
	}

	protected List<MetadataProcessor> metadataProcessors = DEFAULT_PROCESSORS;

	/**
	 * Subclasses should call this to contribute a single component. Their
	 * implementation of contribute() should normally loop over calling this
	 * method. <br>
	 * Can also be called directly.
	 * 
	 * @param componentName name of the component
	 * @param componentClass component class
	 * @param dum DeploymentUnitMetadata to which this component relates
	 */
	public Jsr250Metadata findComponentMetadata(String componentName, Class<?> componentClass,
			DeploymentUnitMetadata dum) {
		Jsr250Metadata jsr250Metadata = newJsr250Metadata(componentName, componentClass, dum);
		// start processing
		if (metadataProcessors != null)
			for (MetadataProcessor processor : metadataProcessors) {
				processor.process(jsr250Metadata);
			}

		return jsr250Metadata;
	}

	/**
	 * Subclasses can override this to return an Jsr250Metadata subclass of
	 * their choice
	 * 
	 * @param componentName component name
	 * @param componentClass class of the component
	 * @param dum deployment unit metadata
	 * @return new instance of Jsr250Metadata or a subclass
	 */
	public Jsr250Metadata newJsr250Metadata(String componentName, Class<?> componentClass, DeploymentUnitMetadata dum) {
		return new TransactionalMetadata(dum, componentName, componentClass);
	}

	/**
	 * Metadata processor chain setter.
	 * 
	 * @param processors
	 */
	public void setMetadataProcessors(List<MetadataProcessor> processors) {
		this.metadataProcessors = processors;
	}

}
