/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.naming.NamingException;

import org.springframework.aop.aspectj.annotation.AspectJAdvisorFactory;
import org.springframework.aop.aspectj.annotation.ReflectiveAspectJAdvisorFactory;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.jee.config.AutodetectBeanFactoryServiceEnvironment;
import org.springframework.jee.config.ServiceEnvironment;
import org.springframework.jee.intercept.InterceptorMetadata;
import org.springframework.jee.validation.Platform;
import org.springframework.jndi.JndiTemplate;

/**
 * Metadata associated with whole deployment unit, such as an EJB JAR file. It
 * is possible to find metadata about all deployed components from this.
 * 
 * @author Rod Johnson
 */
public class DeploymentUnitMetadata {

	private static final String IN_CONTAINER_JNDI_PREFIX = "java:comp/env/";

	/** Used for lookups */
	private JndiTemplate jndiTemplate = new JndiTemplate();

	private AspectJAdvisorFactory aspectJAdvisorFactory = new ReflectiveAspectJAdvisorFactory();

	private List<InterceptorMetadata> defaultInterceptorMetadata = new LinkedList<InterceptorMetadata>();

	private List<Jsr250Metadata> deployedComponents = new LinkedList<Jsr250Metadata>();

	private boolean limitToSpec;

	private List<Class<?>> containerControlInterfaces = new LinkedList<Class<?>>();

	private Map<Class<? extends Exception>, ApplicationExceptionInfo> applicationExceptions = new HashMap<Class<? extends Exception>, ApplicationExceptionInfo>();

	private ServiceEnvironment environment;

	/**
	 * Application context associated with the deployment unit
	 */
	private ApplicationContext deploymentUnitContext;

	private BeanDefinitionRegistry beanDefinitionRegistry;

	private Platform targetPlatform = Platform.JEE_5_0;

	/**
	 * Return the target platform for this deployment unit. Controls validation
	 * rules that will be applied.
	 * 
	 * @return the target platform for this deployment unit.
	 */
	public Platform getTargetPlatform() {
		return this.targetPlatform;
	}

	public void setTargetPlatform(Platform targetPlatform) {
		this.targetPlatform = targetPlatform;
	}

	/**
	 * Associate this DeploymentUnitMetadata with the given deployment unit
	 * level ApplicationContext and corresponding BeanDefinitionRegistry. Must
	 * be called exactly once.
	 * 
	 * @param deploymentContext deployment unit context
	 * @param bdr BeanDefinitionRegistry allowing additions to the context
	 */
	public void associate(ApplicationContext deploymentContext, BeanDefinitionRegistry bdr) {
		if (this.deploymentUnitContext != null) {
			throw new IllegalStateException("Cannot associate DeploymentUnitMetadata with two contexts");
		}
		this.deploymentUnitContext = deploymentContext;
		this.beanDefinitionRegistry = bdr;
	}

	/**
	 * Helper JNDI container-aware lookup method.
	 * 
	 * @param name
	 * @return
	 * @throws NamingException
	 */
	public Object lookup(String name) throws NamingException {
		return jndiTemplate.lookup(composeLookupLocation(name));
	}

	/**
	 * Simple method for composing the final location for a name lookup.
	 * Subclasses should extend this to add custom behavior (i.e. different JNDI prefixes).
	 * @param name
	 * @return
	 */
	public String composeLookupLocation(String name) {
		return IN_CONTAINER_JNDI_PREFIX + name;
	}

	public AspectJAdvisorFactory getAspectJAdvisorFactory() {
		return aspectJAdvisorFactory;
	}

	/**
	 * @return the default interceptors for the deployment unit. Core spec, 11.6
	 */
	public List<InterceptorMetadata> getDefaultInterceptorMetadata() {
		return defaultInterceptorMetadata;
	}

	/**
	 * Add default interceptor metadata to the deploymentUnitMetadata. Useful
	 * for application servers or managed environments.
	 * 
	 * @param defaultIm
	 */
	public void registerDefaultInterceptorMetadata(InterceptorMetadata defaultIm) {
		this.defaultInterceptorMetadata.add(defaultIm);
	}

	/**
	 * @return the deployedComponents
	 */
	public List<Jsr250Metadata> getDeployedComponentMetadata() {
		return this.deployedComponents;
	}

	protected void registerDeployedComponentMetadata(Jsr250Metadata deployedComponentMetadata) {
		this.deployedComponents.add(deployedComponentMetadata);
	}

	/**
	 * @return whether or not the container should limit to the spec; for
	 *         example, not allowing mixing of Spring injection idioms with JEE
	 *         idioms.
	 */
	public boolean isLimitToSpec() {
		return this.limitToSpec;
	}

	/**
	 * @param limitToSpec the limitToSpec to set
	 */
	public void setLimitToSpec(boolean limitToSpec) {
		this.limitToSpec = limitToSpec;
	}

	/**
	 * Register an applicationExceptionInfo. Useful for application servers or
	 * managed environments.
	 * 
	 * @param t applicationException class
	 * @param aei applicationExceptionInfo
	 */
	public void registerApplicationExceptionInfo(Class<? extends Exception> t, ApplicationExceptionInfo aei) {
		applicationExceptions.put(t, aei);
	}

	public ApplicationExceptionInfo getApplicationExceptionInfo(Class<? extends Exception> t) {
		return applicationExceptions.get(t);
	}

	/**
	 * Register the given interface as a container control interface, meaning
	 * that it will not be intercepted
	 * 
	 * @param containerControlInterface container control interface which
	 *            proxies should implement but whose methods will not be
	 *            intercepted
	 */
	public void registerContainerControlInterface(Class<?> containerControlInterface) {
		this.containerControlInterfaces.add(containerControlInterface);
	}

	public List<Class<?>> getContainerControlInterfaces() {
		return containerControlInterfaces;
	}

	/**
	 * Start the deploymentUnitMetada: register the default interceptors and
	 * refresh the deploymentUnitContext as well as metadata contexts.
	 */
	public void startup() {
		// set environment
		// initEnvironment();

		// register default interceptors
		for (InterceptorMetadata defaultIm : defaultInterceptorMetadata) {
			// TODO: is it okay for default interceptors to be singleton?
			RootBeanDefinition rbd = new RootBeanDefinition(defaultIm.getComponentClass(), true);
			defaultIm.setBeanDefinition(rbd);
			// bean definition registration is done automatically by the
			// InterceptorMetadata
			defaultIm.setComponentContext(deploymentUnitContext, beanDefinitionRegistry);
		}

		if ((deploymentUnitContext instanceof AbstractApplicationContext)) {
			((AbstractApplicationContext) this.deploymentUnitContext).refresh();
		}
		else {
			// Assume already started up. Log warning?
		}

		// Start each per-component child container
		for (Jsr250Metadata child : getDeployedComponentMetadata()) {
			child.refresh();
		}
	}

	/**
	 * Used to add a default ServiceEnvironment if none is provided.
	 * 
	 */
	protected void initEnvironment() {
		// TODO: see if we keep this
		if (environment == null) {// use a default one
			environment = new AutodetectBeanFactoryServiceEnvironment(this.deploymentUnitContext);
		}
	}

	/**
	 * @return Returns the applicationExceptions.
	 */
	public Map<Class<? extends Exception>, ApplicationExceptionInfo> getApplicationExceptions() {
		return applicationExceptions;
	}

	/**
	 * @return Returns the environment.
	 */
	public ServiceEnvironment getEnvironment() {
		return environment;
	}

	/**
	 * @param environment The environment to set.
	 */
	public void setEnvironment(ServiceEnvironment environment) {
		this.environment = environment;
	}
}
