/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.util.LinkedList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.jee.validation.InvalidDeploymentUnitException;
import org.springframework.jee.validation.ValidationService;

/**
 * Object that can enrich a GenericApplicationContext with JEE metadata.
 * 
 * @author Rod Johnson
 *
 */
public class Enricher {

	protected final Log log = LogFactory.getLog(getClass());
	
	/**
	 * Metadata about each component deployed to the container.
	 */
	private List<Jsr250Metadata> deployments = new LinkedList<Jsr250Metadata>();

	/**
	 * Application context corresponding to this deployment unit.
	 * Parent of the application contexts for each deployed component.
	 */
	private final ApplicationContext deploymentContext;
	
	/**
	 * Object used to validate the deployment unit before container startup.
	 */
	private ValidationService validationService;
	
	/**
	 * BeanDefinitionRegistry allowing singleton definitions to be retrieved from and added
	 * to the deployment context
	 */
	private final BeanDefinitionRegistry beanDefinitionRegistry;

	private DeploymentUnitMetadata deploymentUnitMetadata;
	
	/**
	 * Create a new Enricher that will act on the given application
	 * context, which relates to a deployment unit such as an
	 * EJB JAR file
	 * @param deploymentContext Spring application context associated
	 * with the deployment unit
	 * @param beanDefinitionRegistry registry allowing singleton definitions to be
	 * manipulated in the deployment unit's context. Can be the same
	 * context if it also implements BeanDefinitionRegistry, otherwise
	 * the BeanDefinitionRegistry associated with it
	 */
	public Enricher(ApplicationContext deploymentContext, BeanDefinitionRegistry beanDefinitionRegistry) {
		this.deploymentContext = deploymentContext;
		this.beanDefinitionRegistry = beanDefinitionRegistry;
		setDeploymentUnitMetadata(new DeploymentUnitMetadata());
	}
	
	
	/**
	 * Convenient method to start up from a GenericApplicationContext,
	 * which implements the BeanDefintionRegistry interface as well as the
	 * ApplicationContext interface
	 * @param deploymenContext application context representing the
	 * deployment unit, which also directly allows BeanDefinition retrieval
	 * and addition
	 */
	public Enricher(GenericApplicationContext deploymentContext) {
		this(deploymentContext, deploymentContext);
	}
	
	/**
	 * Optionally override the default DeploymentUnitMetadata, to
	 * customize JNDI lookup etc.
	 * @param deploymentUnitMetadata new DefaultDeploymentUnitMetadata
	 * to use for components
	 */
	public void setDeploymentUnitMetadata(DeploymentUnitMetadata deploymentUnitMetadata) {
		this.deploymentUnitMetadata  = deploymentUnitMetadata;
		deploymentUnitMetadata.associate(deploymentContext, beanDefinitionRegistry);
	}
	
	/**
	 * @return the deploymentUnitMetadata
	 */
	public DeploymentUnitMetadata getDeploymentUnitMetadata() {
		return this.deploymentUnitMetadata;
	}
	
	/**
	 * Set the optional ValidationService. If this is non-null,
	 * validation will be invoked automatically before trying to start
	 * up the container
	 * @param validationService the validationService to set
	 */
	public void setValidationService(ValidationService validationService) {
		this.validationService = validationService;
	}

	

	/**
	 * Attach the following metadata, that may relate to an existing Spring singleton or a new
	 * component
	 * @param jsr250 JEE metadata
	 * @return the relevant BeanDefinition, which may be an existing Spring-managed singleton,
	 * or wholly new
	 */
	public AbstractBeanDefinition attach(Jsr250Metadata jsr250) {		
		AbstractBeanDefinition bd;
		if (deploymentContext.containsBean(jsr250.getComponentName())) {
			// There's already such a Spring singleton definition.
			// We obtain the definition and enrich it by adding additional metdata
			bd = (AbstractBeanDefinition) beanDefinitionRegistry
					.getBeanDefinition(jsr250.getComponentName());				
			//bd.setSingleton(false);
		} 
		else {
			// Non-singleton
			bd = new RootBeanDefinition(jsr250.getComponentClass(), false);
			beanDefinitionRegistry.registerBeanDefinition(jsr250.getComponentName(), bd);
		}

		GenericApplicationContext componentContext = new GenericApplicationContext(
				deploymentContext);
		componentContext.registerBeanDefinition(Jsr250MetadataBeanPostProcessor.class
				.getName(), new RootBeanDefinition(
				Jsr250MetadataBeanPostProcessor.class));
		
		jsr250.setBeanDefinition(bd);
		jsr250.setComponentContext(componentContext, componentContext);
		
		this.deployments.add(jsr250);		
		return bd;
	}

	
	/**
	 * Start the container. 
	 * @return the container (the same as passed into the
	 * constructor)
	 * @throws InvalidDeploymentUnitException if the metadata is invalid.
	 * This error will be identified before startup.
	 */
	public void startup() throws InvalidDeploymentUnitException {
		if (validationService != null) {
			validationService.validate(deploymentUnitMetadata);
		}
		
		deploymentUnitMetadata.startup();
		
//		log.info("************* Deployment complete: Deployed components with names [" + 
//		StringUtils.collectionToCommaDelimitedString(componentDeployments.keySet()) + "]");
	}

}
