/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.inject;

import java.lang.reflect.Member;

/**
 * Class modelling an injected member of a JEE component.
 * 
 * @author Rod Johnson
 * @author Costin Leau
 */
public abstract class Injection<E extends InjectionInfo> {

	protected final E info;

	protected final Member member;

	private Object value;

	private boolean valueAssigned;

	public E getInfo() {
		return info;
	}

	public Injection(Member member, E resourceInfo) {
		this.info = resourceInfo;
		this.member = member;
	}

	public Object getValue() {
		return value;
	}

	/**
	 * If this method is called, we are dealing with a literal value. A lookup
	 * will not be required: the value will be set.
	 * 
	 * @param value
	 */
	public void setValue(Object value) {
		valueAssigned = true;
		this.value = value;
	}

	public boolean containsValue() {
		return valueAssigned;
	}

	public String getName() {
		return (info.getName() != null ? info.getName() : getDefaultName());
	}

	public Class getType() {
		return (info.getType() != null ? info.getType() : getMemberType());
	}

	/**
	 * Return the default name to be used if the info does not contain one.
	 * 
	 * @return
	 */
	protected String getDefaultName() {
		return member.getName();
	}

	@Override
	public String toString() {
		return getClass().getName() + "; member=" + member.getName() + "; " + info;
	}

	/**
	 * Need to be implemented based on the member type.
	 * 
	 * @return
	 */
	protected abstract Class getMemberType();

	/**
	 * Applies the given value on the instance according to the injection
	 * information - implementation specific.
	 * 
	 * @param instance
	 * @param value
	 */
	public abstract void apply(Object instance, Object value);

}