/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.inject;

import java.lang.reflect.Method;
import java.util.Collections;
import java.util.List;

import javax.interceptor.ExcludeClassInterceptors;
import javax.interceptor.ExcludeDefaultInterceptors;
import javax.interceptor.Interceptors;
import javax.interceptor.InvocationContext;

import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.intercept.InterceptionMetadata;
import org.springframework.jee.intercept.InterceptorMetadata;
import org.springframework.util.ReflectionUtils;

/**
 * EJB3-style interception processor. It will create new bean definitions for each interceptor found (which are
 * later on processed for jsr250 annotations). 
 * 
 * @see org.springframework.jee.inject.Jsr250Processor
 * @see org.springframework.jee.intercept.InterceptionMetadata
 * 
 * @author Rod Johnson
 * @author Costin Leau
 * 
 */
public class InterceptionProcessor implements MetadataProcessor {

	private Jsr250Processor jsr250Processor;

	public InterceptionProcessor() {
		jsr250Processor = new Jsr250Processor();
	}

	public InterceptionProcessor(Jsr250Processor jsr250Processor) {
		this.jsr250Processor = jsr250Processor;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.inject.AnnotationProcessor#process(org.springframework.jee.inject.Jsr250Metadata)
	 */
	public void process(Jsr250Metadata jsr250) {
		if (jsr250 instanceof InterceptionMetadata)
			addInterceptionMetadata((InterceptionMetadata) jsr250);
	}

	/**
	 * Parse the bean and add interception metadata.
	 * 
	 * @param metadata
	 */
	protected void addInterceptionMetadata(InterceptionMetadata metadata) {
		Interceptors interceptorAnnotation = (Interceptors) metadata.getComponentClass().getAnnotation(
				Interceptors.class);

		// class level interceptor
		if (interceptorAnnotation != null) {
			for (Class interceptorClass : interceptorAnnotation.value()) {
				addInterceptorMetadata(metadata, interceptorClass, null);
			}
		}
		// method level interceptors
		for (Class businessInterface : metadata.getBusinessInterfaces()) {
			for (Method methodOnBusinessInterface : businessInterface.getMethods()) {
				// For each method on each business interface
				Method methodOnClass = null;
				try {
					methodOnClass = metadata.getComponentClass().getMethod(methodOnBusinessInterface.getName(),
							(Class[]) methodOnBusinessInterface.getParameterTypes());
				}
				catch (SecurityException ex) {
					// TODO
					throw new UnsupportedOperationException();
				}
				catch (NoSuchMethodException ex) {
					ReflectionUtils.handleReflectionException(ex);
				}
				ExcludeClassInterceptors excludeClassInterceptors = JeeAnnotationUtils.findMethodAnnotation(
						ExcludeClassInterceptors.class, methodOnClass, metadata.getComponentClass());
				ExcludeDefaultInterceptors excludeDefaultInterceptors = JeeAnnotationUtils.findMethodAnnotation(
						ExcludeDefaultInterceptors.class, methodOnClass, metadata.getComponentClass());

				// see the NB below
				if (excludeClassInterceptors != null)
					metadata.setExcludeClassInterceptors(methodOnBusinessInterface);
				if (excludeDefaultInterceptors != null)
					metadata.setExcludeDefaultInterceptors(methodOnBusinessInterface);

				// Find any interceptors annotation
				interceptorAnnotation = JeeAnnotationUtils.findMethodAnnotation(Interceptors.class, methodOnClass,
						metadata.getComponentClass());
				if (interceptorAnnotation != null) {
					for (Class interceptorClass : interceptorAnnotation.value()) {
						// NB: need to match method on business interface, NOT
						// singleton class,
						// as invocation through the proxy will go through the
						// business interface
						addInterceptorMetadata(metadata, interceptorClass, methodOnBusinessInterface);
					}
				}
				metadata.setSelfInterceptorMethod(JeeAnnotationUtils.findAroundInvokeMethod(metadata.getComponentClass()));
			}
		}
	}

	/**
	 * Add the discovered InterceptorMetadata to the given Interception
	 * Metadata.
	 * 
	 * @param componentMetadata
	 * @param interceptorClass
	 *            interceptor class
	 * @param matchingMethod
	 *            matching methdod - can be null which means we have a class
	 *            interceptor
	 */
	protected void addInterceptorMetadata(InterceptionMetadata componentMetadata, Class interceptorClass,
			Method matchingMethod) {

		componentMetadata.registerInterceptorMetadata(createInterceptorMetadata(interceptorClass, matchingMethod));
	}

	/**
	 * Create Interceptor metadata from the given interceptor class.
	 * 
	 * @param interceptorClass
	 * @param matchingMethod
	 * @return
	 */
	public InterceptorMetadata createInterceptorMetadata(Class interceptorClass, Method matchingMethod) {
		// There should be a distinct instance of each interceptor for each
		// singleton instance - using prototype
		RootBeanDefinition interceptorBeanDefinition = new RootBeanDefinition(interceptorClass, false);

		// interceptorBeanDefinition.setSingleton(false);
		InterceptorMetadata im = new InterceptorMetadata(interceptorClass,
				JeeAnnotationUtils.findAroundInvokeMethod(interceptorClass), matchingMethod);
		im.setBeanDefinition(interceptorBeanDefinition);
		jsr250Processor.process(im);

		// find lifecycle methods
		for (LifecycleEvent le : LifecycleEvent.values()) {
			List<Method> listenerMethods = JeeAnnotationUtils.findAnnotatedMethods(im.getComponentClass(), le.getAnnotation(),
					InvocationContext.class);
			// invert order (super classes first)
			Collections.reverse(listenerMethods);
			
			for (Method lifecycleMethod : listenerMethods) {
				im.registerLifecycleEventListenerMethod(le, lifecycleMethod);
			}
		}

		if (matchingMethod == null)
			im.setClassInterceptor(true);

		return im;
	}

}
