/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.naming.NamingException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.beans.factory.ListableBeanFactory;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.AbstractBeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.util.StringUtils;

/**
 * Metadata for JEE 5 injection. Can be supplemented by component-specific
 * metadata such as EJB metadata. Each JeeMetadata is associated with a Spring
 * BeanDefinition and attached as a custom attribute
 * 
 * @see BeanDefinition
 * @author Rod Johnson
 */
public class Jsr250Metadata {

	public static final String KEY = Jsr250Metadata.class.getName();

	protected final Log log = LogFactory.getLog(getClass());

	/**
	 * Spring BeanDefinition this definition belongs to
	 */
	private AbstractBeanDefinition beanDefinition;

	/**
	 * Map from lifecycle event enumeration to callback method defined on a JEE
	 * component class. If there is no such callback method there will be no
	 * entry in the map.
	 */
	private Map<LifecycleEvent, List<Method>> lifecycleCallbackMethods = new HashMap<LifecycleEvent, List<Method>>();

	/**
	 * Name of the JEE component.
	 */
	private final String componentName;

	private List<Injection<?>> injections = new ArrayList<Injection<?>>();

	/**
	 * Spring application context associated with this component. Normally a
	 * child of the deployment unit application context.
	 */
	private ApplicationContext componentContext;

	/**
	 * BeanDefinitionRegistry associated with the component context
	 */
	private BeanDefinitionRegistry beanDefinitionRegistry;

	private final Class<?> componentClass;

	private final DeploymentUnitMetadata deploymentUnitMetadata;

	/**
	 * Copy constructor. Performs a shallow copy.
	 * 
	 * @param other other metadata object to copy field state from
	 */
	protected Jsr250Metadata(Jsr250Metadata other) {
		this(other.deploymentUnitMetadata, other.componentName, other.componentClass);
		this.beanDefinition = other.beanDefinition;
		this.lifecycleCallbackMethods = other.lifecycleCallbackMethods;
		this.componentContext = other.componentContext;
		this.beanDefinitionRegistry = other.beanDefinitionRegistry;
		this.injections = other.injections;
	}

	/**
	 * Construct new component metadata
	 * 
	 * @param deploymentUnitMetadata metadata for whole deployment. This
	 *            metadata will be added to indicate the new component to be
	 *            deployed
	 * @param name name of the component
	 * @param componentClass class of the component
	 */
	public Jsr250Metadata(DeploymentUnitMetadata deploymentUnitMetadata, String name, Class<?> componentClass) {
		this.componentName = name;
		this.componentClass = componentClass;
		this.deploymentUnitMetadata = deploymentUnitMetadata;
		deploymentUnitMetadata.registerDeployedComponentMetadata(this);
	}

	public String getComponentName() {
		return componentName;
	}

	public Class<?> getComponentClass() {
		return componentClass;
	}

	// TODO change name from setter
	public void setComponentContext(ApplicationContext beanFactoryForInstance, BeanDefinitionRegistry bdr) {
		this.componentContext = beanFactoryForInstance;
		this.beanDefinitionRegistry = bdr;
	}

	public final void setComponentContext(GenericApplicationContext gac) {
		this.componentContext = gac;
		this.beanDefinitionRegistry = gac;
	}

	public void setBeanDefinition(AbstractBeanDefinition bd) {
		this.beanDefinition = bd;
		bd.setAttribute(Jsr250Metadata.KEY, this);
		// Override BeanDefinition class if it's not already set
		if (!bd.hasBeanClass()) {
			log.info("Class not specified in Spring configuration for '" + getComponentName()
					+ "': taking from JEE metadata");
			bd.setBeanClass(getComponentClass());
		}
	}

	protected BeanDefinitionRegistry getBeanDefinitionRegistry() {
		return this.beanDefinitionRegistry;
	}

	public ApplicationContext getComponentContext() {
		return componentContext;
	}

	public void registerLifecycleEventCallbackMethod(LifecycleEvent le, Method m) {
		if (m != null) {
			if (!m.isAccessible()) {
				m.setAccessible(true);
			}
			List<Method> methods = lifecycleCallbackMethods.get(le);
			if (methods == null)
				methods = new ArrayList<Method>();

			methods.add(m);
			lifecycleCallbackMethods.put(le, methods);
			log.info("Component lifecycle method: Registered " + m + " on " + getClass().getName() + " for " + le);
		}
	}

	/**
	 * Null if childPostConstruct annotation not found
	 */
	public List<Method> getLifecycleEventCallbackMethod(LifecycleEvent le) {
		return lifecycleCallbackMethods.get(le);
	}

	/**
	 * Return the Spring singleton definition that this additional metadata
	 * relates to
	 * 
	 * @return
	 */
	public AbstractBeanDefinition getBeanDefinition() {
		return beanDefinition;
	}

	public void addInjection(Injection<?> injection) {
		this.injections.add(injection);
	}

	public List<Injection<?>> getInjections() {
		return injections;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(getClass().getSimpleName());
		sb.append("; hashCode=" + hashCode());
		sb.append("; componentName='" + componentName + "'");
		sb.append(": injections=");
		sb.append(StringUtils.collectionToCommaDelimitedString(injections));
		sb.append("; beanDefinition=" + beanDefinition);
		sb.append("; componentContext=" + componentContext);
		return sb.toString();
	}

	/**
	 * Perform dependency injection on the given instance and invoke the
	 * PostConstruct method if any
	 * 
	 * @param instance newly instantiated bean instance. If it is null, static
	 *            members will be used.
	 */
	public void inject(Object instance) {
		applyInjections(instance, injections);
	}

	/**
	 * Apply the given injection list into the passed object.
	 * 
	 * @param instance
	 * @param injections
	 */
	protected void applyInjections(Object instance, List<? extends Injection<?>> injections) {
		for (Injection<?> injection : injections) {
			try {
				Object resolved = injection.containsValue() ? injection.getValue() : resolve(injection);
				log.info("Performing " + injection.getInfo().getInjectionName() + " injection on component with name '" + componentName + "'; "
						+ "Resolved value is of class" + resolved.getClass().getName());
				injection.apply(instance, resolved);
			}
			catch (NoSuchBeanDefinitionException ex) {
				// TODO fix exception
				throw new UnsupportedOperationException("Cannot resolve injection: " + injection, ex);
			}
		}
	}

	/**
	 * Perform dependency injection on the given instance and invoke the
	 * PostConstruct method if any
	 * 
	 * @param instance newly instantiated singleton instance
	 */
	public void injectAndPostConstruct(Object instance) {
		inject(instance);
		invokePostConstructAndRegisterShutdownHook(instance);
	}

	protected void invokePostConstructAndRegisterShutdownHook(Object instance) {
		invokeLifecycleMethods(instance, LifecycleEvent.POST_CONSTRUCT);
		// TODO register destroy method being childPreDestroy
	}

	/**
	 * Resolve the given injected value Look first in the Spring singleton
	 * factory, then in JNDI name
	 * 
	 * @param injection injection metadata
	 * @return object from Spring context or JNDI
	 */
	protected Object resolve(Injection injection) {
		try {
			return componentContext.getBean(injection.getName());
		}
		catch (NoSuchBeanDefinitionException ex) {
			try {
				return getDeploymentUnitMetadata().lookup(injection.getName());
			}
			catch (NamingException namingException) {
				log.error("Can't find JNDI location '"
						+ getDeploymentUnitMetadata().composeLookupLocation(injection.getName()) + "'");
			}

			// Try autowire by type
			return resolveByType(injection.getType());
		}
	}

	/**
	 * Subclasses that know about a context can perform special purpose
	 * injection here
	 * 
	 * @param type
	 * @return
	 * @throws NoSuchBeanDefinitionException
	 */
	protected Object resolveByType(Class type) throws NoSuchBeanDefinitionException {
		return getUniqueInstanceOfType(componentContext, type);
	}

	public Object getUniqueInstanceOfType(ListableBeanFactory lbf, Class type) {
		Map beansOfType = BeanFactoryUtils.beansOfTypeIncludingAncestors(lbf, type);
		if (beansOfType.isEmpty()) {
			throw new NoSuchBeanDefinitionException(type, "No beans of type " + type.getName() + "; owner=" + lbf);
		}
		if (beansOfType.size() > 1) {
			throw new NoSuchBeanDefinitionException(type, beansOfType.size() + " beans of type " + type.getName()
					+ ": must be unique to support injection by inferred type");
		}
		// Return the single value
		return beansOfType.values().iterator().next();
	}

	// TODO properly handle exceptions. Translate to EJB exceptions, or whatever
	public void invokeLifecycleMethods(Object bean, LifecycleEvent le) {
		List<Method> lifecyleMethods = getLifecycleEventCallbackMethod(le);
		if (lifecyleMethods != null)
			for (Method method : lifecyleMethods) {
				invokeLifecycleMethod(bean, method);
			}

	}

	protected boolean invokeLifecycleMethod(Object bean, Method lifecyleMethod) {
		return invokeLifecycleMethod(bean, lifecyleMethod, null);
	}

	public boolean invokeLifecycleMethod(Object bean, Method lifecyleMethod, Object[] args) {
		if (lifecyleMethod != null) {
			try {
				lifecyleMethod.invoke(bean, args);
			}
			catch (IllegalArgumentException ex) {
				throw new IllegalStateException("Failure to invoke " + lifecyleMethod + " on " + bean.getClass()
						+ ": args=[" + StringUtils.arrayToCommaDelimitedString(args) + "]", ex);
			}
			catch (IllegalAccessException ex) {
				throw new UnsupportedOperationException(ex.toString());
			}
			catch (InvocationTargetException ex) {
				// TODO fix exception
				throw new UnsupportedOperationException("Lifecycle method on singleton with metadata " + this
						+ " threw exception", ex.getTargetException());
			}
			return true;
		}

		return false;
	}

	public DeploymentUnitMetadata getDeploymentUnitMetadata() {
		return deploymentUnitMetadata;
	}

	public void refresh() {
		if (componentContext instanceof AbstractApplicationContext) {
			((AbstractApplicationContext) componentContext).refresh();
		}
	}

}
