/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.inject;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.List;

import javax.annotation.Resource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.util.ReflectionUtils;

/**
 * Jsr250 annotation processor. Creates field and method injection based on the
 * annotations found.
 * 
 * @see org.springframework.jee.inject.FieldInjection
 * @see org.springframework.jee.inject.MethodInjection
 * 
 * @author Costin Leau
 * 
 */
public class Jsr250Processor implements MetadataProcessor {

	protected final Log log = LogFactory.getLog(getClass());

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.inject.AnnotationProcessor#process(org.springframework.jee.inject.Jsr250Metadata)
	 */
	public void process(Jsr250Metadata im) {
		addInjectionMetadata(im);
		addLifecycleMethods(im);
	}

	protected void addInjectionMetadata(Jsr250Metadata jsr250) {
		addAnnotationClassLevel(jsr250);
		addAnnotationFieldInjections(jsr250);
		addAnnotationMethodInjections(jsr250);
	}

	/**
	 * Add support for class level annotations like
	 * 
	 * @Resources
	 * 
	 * @param jsr250
	 */
	private void addAnnotationClassLevel(Jsr250Metadata jsr250) {
		// throw new UnsupportedOperationException();
	}

	private void addAnnotationFieldInjections(final Jsr250Metadata jem) {
		ReflectionUtils.doWithFields(jem.getComponentClass(), new ReflectionUtils.FieldCallback() {
			public void doWith(Field f) throws IllegalArgumentException, IllegalAccessException {
				processCandidateInjectionField(jem, f);
			}
		});
	}

	private void processCandidateInjectionField(Jsr250Metadata jem, Field f) {
		Resource resource = f.getAnnotation(Resource.class);
		if (resource != null) {
			FieldInjection fi = new FieldInjection(f, resource);
			jem.addInjection(fi);
		}
	}

	private void addAnnotationMethodInjections(final Jsr250Metadata jem) {
		ReflectionUtils.doWithMethods(jem.getComponentClass(), new ReflectionUtils.MethodCallback() {
			public void doWith(Method m) throws IllegalArgumentException, IllegalAccessException {
				processCandidateInjectionMethod(jem, m);
			}
		});
	}

	private void processCandidateInjectionMethod(Jsr250Metadata jem, Method m) {
		Resource resource = (Resource) JeeAnnotationUtils.findMethodAnnotation(Resource.class, m,
				jem.getComponentClass());
		if (resource != null) {
			MethodInjection mi = new MethodInjection(m, resource);
			jem.addInjection(mi);
		}
	}

	protected void addAnnotationResourceReferences(List<Injection> injections, Class clazz) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Find lifecycle methods
	 * 
	 * @param jsr250 JSR 250 metadata already built
	 */
	private void addLifecycleMethods(Jsr250Metadata jsr250) {
		for (LifecycleEvent le : LifecycleEvent.values()) {
			// find only bean methods - interceptor methods will be registered
			// separately.
			List<Method> methods = JeeAnnotationUtils.findAnnotatedMethods(jsr250.getComponentClass(),
					le.getAnnotation(), new Class[0]);
			// invert order (superclasses first)
			Collections.reverse(methods);
			for (Method method : methods) {
				jsr250.registerLifecycleEventCallbackMethod(le, method);
			}
		}
	}
}
