/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.lang.reflect.Method;

import javax.ejb.Stateful;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.transaction.TransactionalMetadata;
import org.springframework.util.ReflectionUtils;

/**
 * Ejb3 transaction annotation processor.
 * 
 * @author Costin Leau
 * 
 */
public class TransactionProcessor implements MetadataProcessor {

	private static final Log log = LogFactory.getLog(TransactionProcessor.class);

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.inject.MetadataProcessor#process(org.springframework.jee.inject.Jsr250Metadata)
	 */
	public void process(Jsr250Metadata jsr250) {
		if (jsr250 instanceof TransactionalMetadata)
			addTransactionalMetadata((TransactionalMetadata) jsr250);
	}

	protected void addTransactionalMetadata(TransactionalMetadata metadata) {
		boolean debug = log.isDebugEnabled();

		// analyze class methods
		Class clazz = metadata.getComponentClass();

		if (isClassTransactional(clazz)) {
			if (debug)
				log.debug("adding transactional metadata...");
			metadata.setTransactional(true);
			addTransactionalMethodMetadata(metadata);
		}
	}

	@SuppressWarnings("unchecked")
	protected boolean isClassTransactional(Class clazz) {
		boolean debug = log.isDebugEnabled();

		if (debug) {
			log.debug("analyzing class " + clazz);
		}
		// see if it's stateless or stateful
		boolean isSessionBean = (clazz.isAnnotationPresent(Stateless.class) || clazz.isAnnotationPresent(Stateful.class));

		// see if it manages its own transaction or not
		if (isSessionBean) {
			if (debug)
				log.debug(clazz + " is a session singleton");
			TransactionManagement tmAnnotation = (TransactionManagement) clazz.getAnnotation(TransactionManagement.class);
			if (tmAnnotation != null) {
				if (tmAnnotation.value() == TransactionManagementType.BEAN) {
					if (debug)
						log.debug(clazz + " has singleton management transaction.");
					return false;
				}
			}
			if (debug)
				log.debug(clazz + " has container management transaction");
			return true;
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	protected void addTransactionalMethodMetadata(TransactionalMetadata metadata) {
		boolean debug = log.isDebugEnabled();
		// get methods
		for (Class businessInterface : metadata.getBusinessInterfaces()) {
			Method methodOnClass = null;
			for (Method methodOnBusinessInterface : businessInterface.getMethods()) {
				try {
					methodOnClass = metadata.getComponentClass().getMethod(methodOnBusinessInterface.getName(),
							(Class[]) methodOnBusinessInterface.getParameterTypes());

					// for each method
					// 1. get tx attributes
					TransactionAttribute txAttribute = JeeAnnotationUtils.findInheritedMemberAnnotation(
							TransactionAttribute.class, metadata.getComponentClass(), methodOnClass);
					if (debug) {
						log.debug("adding txAttribute " + txAttribute + " to method " + methodOnClass);
					}
					metadata.addTransactionalMethod(methodOnClass, txAttribute);
					// 2. get checked exceptions
					for (Class<?> exception : methodOnClass.getExceptionTypes()) {
						// look only at checked exceptions
						if (!RuntimeException.class.isAssignableFrom(exception) && !Error.class.isAssignableFrom(exception)) {
							metadata.addApplicationCheckedException((Class<? extends Exception>) exception);
						}
						if (debug) {
							log.debug("adding application exception " + exception);
						}
					}
				}

				catch (SecurityException ex) {
					// TODO : translate this into a DeploymentUnit exception
					throw new UnsupportedOperationException();
				}
				catch (NoSuchMethodException ex) {
					ReflectionUtils.handleReflectionException(ex);
				}
			}
		}
	}
}
