/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.ejb.EJBContext;
import javax.ejb.SessionSynchronization;
import javax.interceptor.AroundInvoke;
import javax.interceptor.InvocationContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.aop.Advisor;
import org.springframework.aop.aspectj.annotation.AnnotationAwareAspectJAutoProxyCreator;
import org.springframework.aop.framework.Advised;
import org.springframework.aop.framework.AopContext;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aop.interceptor.ExposeInvocationInterceptor;
import org.springframework.aop.support.AopUtils;
import org.springframework.aop.support.DelegatingIntroductionInterceptor;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.config.BeanFactoryPostProcessor;
import org.springframework.beans.factory.config.BeanPostProcessor;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.context.ApplicationContext;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.jee.inject.LifecycleEvent;
import org.springframework.util.ClassUtils;

/**
 * Class representing metadata for a component that may have interceptors as per
 * the JEE (EJB 3.0) interception specification. Attached to a Spring singleton
 * definition.
 * 
 * @author Rod Johnson
 */
public class InterceptionMetadata extends Jsr250Metadata {

	private static final Log log = LogFactory.getLog(InterceptionMetadata.class);

	/**
	 * Interfaces that cannot be considered as business interfaces if we
	 * autodetect business interfaces
	 */
	private static Set<Class> EXCLUDED_BUSINESS_INTERFACES = new HashSet<Class>();
	static {
		EXCLUDED_BUSINESS_INTERFACES.add(Cloneable.class);
		EXCLUDED_BUSINESS_INTERFACES.add(Serializable.class);
		EXCLUDED_BUSINESS_INTERFACES.add(SessionSynchronization.class);
		EXCLUDED_BUSINESS_INTERFACES.add(InitializingBean.class);
		EXCLUDED_BUSINESS_INTERFACES.add(BeanPostProcessor.class);
		EXCLUDED_BUSINESS_INTERFACES.add(BeanFactoryPostProcessor.class);
		EXCLUDED_BUSINESS_INTERFACES.add(DisposableBean.class);
	};

	/**
	 * Class representing the exclusion of class or default interceptors from a
	 * particular method
	 * 
	 * @author Costin Leau
	 */
	public static class InterceptorExclusion {

		private boolean excludeClassInterceptors;

		private boolean excludeDefaultInterceptors;

		/**
		 * @return the excludeClassInterceptors
		 */
		public boolean isExcludeClassInterceptors() {
			return excludeClassInterceptors;
		}

		/**
		 * @param excludeClassInterceptors
		 *            the excludeClassInterceptors to set
		 */
		public void setExcludeClassInterceptors(boolean excludeClassInterceptors) {
			this.excludeClassInterceptors = excludeClassInterceptors;
		}

		/**
		 * @return the excludeDefaultInterceptors
		 */
		public boolean isExcludeDefaultInterceptors() {
			return excludeDefaultInterceptors;
		}

		/**
		 * @param excludeDefaultInterceptors
		 *            the excludeDefaultInterceptors to set
		 */
		public void setExcludeDefaultInterceptors(boolean excludeDefaultInterceptors) {
			this.excludeDefaultInterceptors = excludeDefaultInterceptors;
		}
	}

	/**
	 * Exclusion list for each method.
	 */
	private Map<Method, InterceptorExclusion> exclusions = new HashMap<Method, InterceptorExclusion>();

	/**
	 * Class and method level interception
	 */
	private List<InterceptorMetadata> interceptorMetadataList = new LinkedList<InterceptorMetadata>();

	private Method selfInterceptorMethod;

	/**
	 * For explicit interceptor order declared in the interceptor-order in the
	 * DD List of object that interceptor-order is declared on, this can be
	 * either the bean class or a method.
	 */
	private Set<Object> interceptorOrderObject = new HashSet<Object>();

	private Map<Object, List<InterceptorMetadata>> interceptorOrderMap = new HashMap<Object, List<InterceptorMetadata>>();

	/**
	 * The business methods on the singleton
	 */
	private List<Class<?>> businessInterfaces;

	public InterceptionMetadata(DeploymentUnitMetadata dum, String name, Class<?> componentClass) {
		super(dum, name, componentClass);
	}

	/**
	 * Set the application context for the superclass and for the inner
	 * interceptorMetadata.
	 */
	@Override
	public void setComponentContext(ApplicationContext componentContext, BeanDefinitionRegistry bdr) {
		super.setComponentContext(componentContext, bdr);
		for (Iterator<List<InterceptorMetadata>> i = interceptorOrderMap.values().iterator(); i.hasNext();) {
			for (InterceptorMetadata im : i.next()) {
				im.setComponentContext(componentContext, bdr);
			}
		}
		for (InterceptorMetadata im : interceptorMetadataList) {
			im.setComponentContext(componentContext, bdr);
		}

		// createBeanDefinitionsForInterceptors();
	}

	@Override
	protected void invokePostConstructAndRegisterShutdownHook(Object instance) {
		// Suppress superclass implementation. Do nothing.
	}

	/**
	 * Set the self interceptor method for this component
	 * 
	 * @param selfInterceptorMethod
	 *            the self interceptor method for this component
	 */
	public void setSelfInterceptorMethod(Method selfInterceptorMethod) {
		this.selfInterceptorMethod = selfInterceptorMethod;
	}

	/**
	 * Add a business interface for the singleton. If no business methods are
	 * added, default to using introspection to find all interfaces on the class
	 * 
	 * @param businessInterface
	 *            business interface to add
	 */
	public void addBusinessInterface(Class<?> businessInterface) {
		this.businessInterfaces.add(businessInterface);
	}

	public void setExcludeClassInterceptors(Method method) {
		InterceptorExclusion exclude = exclusions.get(method);
		if (exclude == null) {
			exclude = new InterceptorExclusion();
		}
		exclude.setExcludeClassInterceptors(true);
		exclusions.put(method, exclude);
	}

	public void setExcludeDefaultInterceptors(Method method) {
		InterceptorExclusion exclude = exclusions.get(method);
		if (exclude == null) {
			exclude = new InterceptorExclusion();
		}
		exclude.setExcludeDefaultInterceptors(true);
		exclusions.put(method, exclude);
	}

	/**
	 * Add interceptor metadata. Whether it's a class or method level
	 * interceptor will be defined by whether the matchingMethod it returns is
	 * null. Following the EJB 3.0 specification, method level interceptors will
	 * be added after all class level interceptors.
	 * 
	 * @param im
	 *            interceptor
	 * @param matchingMethod
	 *            method to match on
	 */
	public void registerInterceptorMetadata(InterceptorMetadata im) {

		if (im.getMatchingMethod() != null) {
			if (log.isDebugEnabled())
				log.debug("registering as method interceptor " + im);

			// It's a method interceptor
			// Add to the end of the list
			this.interceptorMetadataList.add(im);

		}
		else {
			// It's a class interceptor.
			// Add before any method interceptors
			int insertionPos = 0;
			while (insertionPos < interceptorMetadataList.size()
					&& interceptorMetadataList.get(insertionPos).getMatchingMethod() == null) {
				++insertionPos;
			}
			this.interceptorMetadataList.add(insertionPos, im);

			if (log.isDebugEnabled())
				log.debug("registering as class interceptor " + im);

		}
	}

	/**
	 * Return the interceptors that apply to the given component.
	 * 
	 * @return a list of interceptors applying to the given component. This will
	 *         be sorted, according to the spec, as with class interceptors
	 *         followed by methodInterceptor, and finally any around invoke
	 *         method denoting an interceptor on the component
	 */
	public List<InterceptorMetadata> getInterceptorMetadata() {
		return this.interceptorMetadataList;
	}

	public void registerInterceptorOrder(InterceptorMetadata im, Object object) {
		this.interceptorOrderObject.add(object);

		List<InterceptorMetadata> interceptorOrder = interceptorOrderMap.get(object);
		if (interceptorOrder == null)
			interceptorOrder = new LinkedList<InterceptorMetadata>();

		interceptorOrder.add(im);
		interceptorOrderMap.put(object, interceptorOrder);
	}

	public List<InterceptorMetadata> getInterceptorOrder(Object object) {
		return interceptorOrderMap.get(object);
	}

	public List<Class<?>> getBusinessInterfaces() {
		if (this.businessInterfaces == null) {
			this.businessInterfaces = findBusinessInterfacesFromClassOrAnnotation(getComponentClass());
		}
		return this.businessInterfaces;
	}

	private List<Class<?>> findBusinessInterfacesFromClassOrAnnotation(Class<?> clazz) {
		List<Class<?>> businessInterfaces = new LinkedList<Class<?>>();
		for (Class<?> potentialBusinessInterface : ClassUtils.getAllInterfacesForClass(clazz)) {
			if (!isExcludedAsBusinessInterfaceOnAutodetect(potentialBusinessInterface)
					&& !getDeploymentUnitMetadata().getContainerControlInterfaces().contains(potentialBusinessInterface)) {
				businessInterfaces.add(potentialBusinessInterface);
			}
		}

		// Exclusions
		businessInterfaces.removeAll(EXCLUDED_BUSINESS_INTERFACES);

		if (log.isDebugEnabled())
			log.debug("for class " + clazz.getName() + " found business interfaces: " + businessInterfaces);

		return businessInterfaces;
	}

	/**
	 * Subclasses can override this to exclude further business interfaces
	 * 
	 * @param potentialBusinessInterface
	 *            candidate interface
	 * @return whether this class should be excluded from consideration as a
	 *         business interface if autodetecting business interfaces
	 */
	protected boolean isExcludedAsBusinessInterfaceOnAutodetect(Class<?> potentialBusinessInterface) {
		return false;
	}

	/**
	 * Use Spring AOP to create a proxy invoking the JEE-style interceptors and
	 * any Spring-style interceptors. Create new instances of the given
	 * interceptors.
	 * 
	 * @param instance
	 *            the singleton instance to wrap in a proxy
	 * @return a Spring AOP proxy-control interface allowing subsequent changes
	 *         to the proxy, if the object was proxied at all
	 */
	public Object createProxyIfNecessary(Object target) {
		return createProxyIfNecessary(target, new HashMap<Class, Object>());
	}

	/**
	 * Use Spring AOP to create a proxy invoking the JEE-style interceptors and
	 * any Spring-style interceptors. Allows the creation of a fresh proxy
	 * wrapping a given target singleton instance and set of interceptors: for
	 * example, in the case where we have passivated a SFSB and now need to
	 * create a new proxy given activated interceptors and target instance
	 * state.
	 * 
	 * @param instance
	 *            the singleton instance to wrap in a proxy
	 * @param interceptorInstancesForClass
	 *            map from interceptor class to instance of JEE interceptors
	 * @return a Spring AOP proxy-control interface allowing subsequent changes
	 *         to the proxy, if the object was proxied at all
	 */
	public Object createProxyIfNecessary(Object target, Map<Class, Object> interceptorInstancesForClass) {
		// Use Spring AOP's programmatic way to create a proxy
		ProxyFactory pf = new ProxyFactory();
		pf.setTarget(target);

		// Don't use CGLIB. We only ever want to expose interfaces.
		pf.setProxyTargetClass(false);

		// We do this to allow reentry according to EJB rules
		pf.setExposeProxy(true);

		// Only proxy the business interfaces
		for (Class businessInterface : getBusinessInterfaces()) {
			pf.addInterface(businessInterface);
		}
		for (Class containerControlInterface : getDeploymentUnitMetadata().getContainerControlInterfaces()) {
			pf.addInterface(containerControlInterface);
		}

		// getDeploymentMetadata().getServiceEnvironment().addSharedAdvisors(pf,
		// this.enforceCallByValueForRemote && isRemote());

		// addSharedAdvisors(pf);

		addComponentContractInterceptors(pf);

		addJeeInterceptors(pf, target, interceptorInstancesForClass);

		// Add Spring advisors to enable Spring AOP if this is allowed
		if (!getDeploymentUnitMetadata().isLimitToSpec()) {
			addSpringAopAdvisors(pf, target);
			addAspectJAdvisors(pf, target);
		}

		customizeProxyConfig(pf);

		if (pf.getAdvisors().length == 0) {
			// Let superclass handle lifecycle methods
			super.invokePostConstructAndRegisterShutdownHook(target);
			return target;
		}
		else {
			// Add proxy control introduction to head of chain
			pf.addAdvice(0, new DelegatingIntroductionInterceptor(new AdvisorChainProxyControl()));
			return pf.getProxy();
		}
	}

	/**
	 * Add interceptors that apply the component contract--such as EJB
	 * transaction behaviour.
	 * 
	 * @param pf
	 *            ProxyFactory used to create proxy
	 */
	protected void addComponentContractInterceptors(ProxyFactory pf) {
		// Do nothing in this class
	}

	/**
	 * Perform optional customization of ProxyFactory before use
	 * 
	 * @param pf
	 */
	protected void customizeProxyConfig(ProxyFactory pf) {
		// Do nothing in this case
	}

	/**
	 * Mixin to allow introduction of the __ProxyControl control interface. Only
	 * works within an AOP invocation context. Walks the proxy chain looking for
	 * MethodInterceptorPointcutAdvisors targeting EJB3-style interceptors.
	 */
	private class AdvisorChainProxyControl implements __ProxyControl {

		public InterceptionMetadata getMetadata() {
			return InterceptionMetadata.this;
		}

		public List<JeeInterceptorPointcutAdvisor> getJeeInterceptorAdvisors() {
			// TODO cache? Assume won't change?
			Advised advised = (Advised) AopContext.currentProxy();

			List<JeeInterceptorPointcutAdvisor> advisors = new ArrayList<JeeInterceptorPointcutAdvisor>(
					advised.getAdvisors().length);
			for (Advisor a : advised.getAdvisors()) {
				if (a instanceof JeeInterceptorPointcutAdvisor) {
					JeeInterceptorPointcutAdvisor mipa = (JeeInterceptorPointcutAdvisor) a;
					advisors.add(mipa);
				}
			}
			return advisors;
		}

		/**
		 * Invoke the given lifecycle event method (if found) on each advisor in
		 * the chain
		 * 
		 * @param le
		 *            lifecycle event
		 */
		public void invokeLifecycleMethod(LifecycleEvent le) {
			// Delegate to the metadata, given the proxy instance
			getMetadata().invokeLifecycleMethods(AopContext.currentProxy(), le);
		}
	}

	/**
	 * Override superclass method to invoke all interceptors in the chain if
	 * they listen to this method
	 * 
	 * @param singleton
	 *            normally a proxy
	 * @param le
	 *            lifecycle event
	 */
	@Override
	public void invokeLifecycleMethods(Object bean, LifecycleEvent le) {
		if (!(bean instanceof __ProxyControl)) {
			super.invokeLifecycleMethods(bean, le);
			return;
		}

		LifecycleEventCallbackInvocationContext lifecycleEventCallbackInvocationContext = new LifecycleEventCallbackInvocationContext(bean);

		for (JeeInterceptorPointcutAdvisor mipa : ((__ProxyControl) bean).getJeeInterceptorAdvisors()) {
			lifecycleEventCallbackInvocationContext.shortCircuited = true;
			InterceptorMetadata im = mipa.getAdvice().getInterceptorMetadata();
			if (im == null) {
				// Probably a self interceptor. No metadata. So
				// no information about the method to invoke. Not an error.
				continue;
			}
			List<Method> listeners = im.getLifecycleEventListenerMethod(le);
			if (listeners != null)
				for (Method lcm : listeners) {
					boolean invoked = mipa.getAdvice().getInterceptorMetadata().invokeLifecycleMethod(
							mipa.getAdvice().getJeeInterceptorInstance(), lcm,
							new Object[] { lifecycleEventCallbackInvocationContext });
					if (invoked && lifecycleEventCallbackInvocationContext.shortCircuited) {
						// Terminate processing of chain due to failure to call
						// proceed
						break;
					}
				}
		}

		// Now invoke on the target (singleton instance) itself
		// TODO Do we do this first? If not short-circuited?
		try {
			bean = ((Advised) bean).getTargetSource().getTarget();
			super.invokeLifecycleMethods(bean, le);
		}
		catch (Exception ex) {
			// TargetSource.getTarget declares that it throws an Exception,
			// but this should be a singleton target so this cannot happen
			log.warn("Illegal state: can't get target", ex);
		}
	}

	/**
	 * Special InvocationContext implementation to use when invoking lifecycle
	 * listenr methods on EJB3-style interceptors.
	 */
	private class LifecycleEventCallbackInvocationContext implements InvocationContext {

		private Map<String, Object> contextData = new HashMap<String, Object>();

		private boolean shortCircuited;
		
		private Object bean;

		public LifecycleEventCallbackInvocationContext(Object bean) {
			this.bean = bean;
		}

		public Object getTarget() {
			Object target = null;
			try {
				target = ((Advised) bean).getTargetSource().getTarget();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			return target;
		}

		public Method getMethod() {
			return null;
		}

		public Object[] getParameters() {
			return null;
		}

		public void setParameters(Object[] params) {
			throw new UnsupportedOperationException("Cannot set parameters on LifecycleEventCallbackInvocationContext");
		}

		public EJBContext getEJBContext() {
			throw new UnsupportedOperationException("TODO: Figure out how to pass this in");
		}

		public Map<String, Object> getContextData() {
			return contextData;
		}

		public Object proceed() throws Exception {
			shortCircuited = false;
			// Can never have arguments
			return null;
		}
	}

	/**
	 * Add EJB3 style interceptors to this singleton instance
	 * 
	 * @param advised
	 *            Spring AOP proxy control interface
	 * @param singleton
	 *            new JEE component instance being advised (intercepted)
	 * @param interceptorInstancesForClass
	 *            map from interceptor class to interceptor instance. Used to
	 *            ensure that we have only one instance of each interceptor
	 *            class, as per Core Contract specification 11.7.
	 */
	protected void addJeeInterceptors(Advised advised, Object bean, Map<Class, Object> interceptorInstancesForClass) {

		if (log.isDebugEnabled())
			log.debug("adding interceptors for " + bean);

		// TODO what are the lifecycle requirements on default interceptor?
		// Seems unclear in spec.
		// Is it tied to the deployment unit's context (seems reasonable)

		// 1. add default interceptors
		addJeeInterceptors(advised, getDeploymentUnitMetadata().getDefaultInterceptorMetadata(),
				interceptorInstancesForClass, false);

		// 2. add declared interceptors (based on metadata (annotations)).
		addJeeInterceptors(advised, interceptorMetadataList, interceptorInstancesForClass, false);

		// 3. add self interceptor
		addSelfInterceptor(advised, bean);

		// 4. add interceptor-order if it exists
		for (Object o : interceptorOrderObject) {
			addJeeInterceptors(advised, interceptorOrderMap.get(o), interceptorInstancesForClass, true);
		}
	}

	/**
	 * Parses the interceptor metadata list, creates Spring pointcut advisors
	 * based on it and adds it to the given advised object.
	 * 
	 * @param advised
	 * @param interceptorMetadata
	 * @param interceptorInstancesForClass
	 */
	protected void addJeeInterceptors(Advised advised, List<InterceptorMetadata> interceptorMetadata,
			Map<Class, Object> interceptorInstancesForClass, boolean isInterceptorOrder) {

		for (InterceptorMetadata im : interceptorMetadata) {

			boolean noMatchingObject = false;
			Set<Object> currentInterceptorOrderObject = new HashSet<Object>();
			if (!isInterceptorOrder) {
				currentInterceptorOrderObject = interceptorOrderObject;
				Object matchingObject = im.getMatchingMethod();
				if (matchingObject == null)
					matchingObject = im.getComponentClass();
				if (!interceptorOrderObject.contains(matchingObject))
					noMatchingObject = true;
			}

			if (isInterceptorOrder || noMatchingObject) {
				// See if we already have an interceptor of this class
				Object interceptorInstance = interceptorInstancesForClass.get(im.getComponentClass());
				if (interceptorInstance == null) {
					interceptorInstance = getComponentContext().getBean(im.getComponentName());
					// Remember that we found this interceptor instance for this
					// class
					interceptorInstancesForClass.put(im.getComponentClass(), interceptorInstance);
				}

				boolean debug = log.isDebugEnabled();

				if (debug)
					log.debug("creating ee interceptor from " + im + " for instance " + interceptorInstance);

				// Create base (AOP Alliance) interceptor
				JeeInterceptorInterceptor jeeii = new JeeInterceptorInterceptor(im, interceptorInstance);
				// Create base interceptor advisor
				JeeInterceptorPointcutAdvisor mipa = new JeeInterceptorPointcutAdvisor(im.getMatchingMethod(), jeeii,
						getDeploymentUnitMetadata(), exclusions, currentInterceptorOrderObject);

				// add super interceptors (before the actual interceptor)
				addSuperInterceptors(advised, interceptorInstance, mipa);

				advised.addAdvisor(mipa);
			}
		}
	}

	protected void addSuperInterceptors(Advised advised, Object baseInterceptorInstance,
			JeeInterceptorPointcutAdvisor baseInterceptorPointcutAdvisor) {

		List<Method> superAroundInvokes = findSuperAroundInvokes(baseInterceptorInstance.getClass());

		if (log.isDebugEnabled() && !superAroundInvokes.isEmpty()) {
			log.debug("creating ee superinterceptor for instance " + baseInterceptorInstance);
		}
		for (Method superInterceptorMethod : superAroundInvokes) {
			JeeInterceptorInterceptor superJeeii = new JeeInterceptorInterceptor(baseInterceptorInstance,
					superInterceptorMethod);
			// Use a delegating pointcut to reuse the base interceptor
			// advisor
			SuperJeeInterceptorPointcutAdvisor sipa = new SuperJeeInterceptorPointcutAdvisor(
					baseInterceptorPointcutAdvisor, superJeeii);
			advised.addAdvisor(sipa);
		}
	}

	protected List<Method> findSuperAroundInvokes(Class clazz) {
		// find super interceptor methods
		List<Method> superAroundInvokes = JeeAnnotationUtils.findAnnotatedMethods(clazz.getSuperclass(),
				AroundInvoke.class, JeeAnnotationUtils.AROUND_INVOKE_METHOD_PARAM);
		// invert list (to get correct call order)
		Collections.reverse(superAroundInvokes);
		return superAroundInvokes;
	}

	/**
	 * Add an interceptor definition for the singleton/bean itself.
	 * 
	 * @param advised
	 *            AOP proxy control interface
	 * @param singleton
	 *            target instance that contains an around invoke method
	 */
	protected void addSelfInterceptor(Advised advised, Object bean) {
		if (selfInterceptorMethod != null) {
			// JeeInterceptorInterceptor jeeii = new
			// JeeInterceptorInterceptor(new InterceptorMetadata(this,
			// selfInterceptorMethod), bean);
			JeeInterceptorInterceptor jeeii = new JeeInterceptorInterceptor(bean, selfInterceptorMethod);

			JeeInterceptorPointcutAdvisor mipa = new JeeInterceptorPointcutAdvisor(null, jeeii,
					getDeploymentUnitMetadata(), this.exclusions, interceptorOrderObject);

			// add super interceptors (before the actual interceptor)
			addSuperInterceptors(advised, bean, mipa);

			advised.addAdvisor(mipa);
		}
	}

	@SuppressWarnings("unchecked")
	protected int addSpringAopAdvisors(Advised pf, Object bean) {
		List<Advisor> advisorsInFactory = new LinkedList<Advisor>();
		String[] advisorNames = BeanFactoryUtils.beanNamesForTypeIncludingAncestors(getComponentContext(),
				Advisor.class, true, false);
		List<Advisor> candidateAdvisors = new LinkedList<Advisor>();
		for (String advisorName : advisorNames) {
			candidateAdvisors.add((Advisor) getComponentContext().getBean(advisorName));
		}

		List<Advisor> eligibleAdvisors = (List<Advisor>) AopUtils.findAdvisorsThatCanApply(advisorsInFactory,
				bean.getClass());
		for (Advisor advisor : eligibleAdvisors) {
			pf.addAdvisor(advisor);
		}
		return eligibleAdvisors.size();
	}

	protected int addAspectJAdvisors(Advised advised, Object bean) {
		AnnotationAwareAspectJAutoProxyCreator proxyCreator = new AnnotationAwareAspectJAutoProxyCreator();
		List<Advisor> eligibleAdvisors = proxyCreator.createAspectJAdvisors(
				getDeploymentUnitMetadata().getAspectJAdvisorFactory(), getComponentContext());
		if (!eligibleAdvisors.isEmpty()) {
			advised.addAdvisor(ExposeInvocationInterceptor.ADVISOR);
		}
		for (Advisor a : eligibleAdvisors) {
			advised.addAdvisor(a);
		}
		return eligibleAdvisors.size();
	}

	/**
	 * @return the exclusion
	 */
	public Map<Method, InterceptorExclusion> getExclusion() {
		return exclusions;
	}

}
