/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.context.ApplicationContext;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.jee.inject.LifecycleEvent;

/**
 * Class representing metadata about an interceptor class. Includes JSR-250
 * metadata that can drive injection and lifecycle callbacks, and identification
 * of an around invoke method on the class. The latter information is often, but
 * not necessarily, supplied using an annotation.
 * 
 * @author Rod Johnson
 * 
 */
public class InterceptorMetadata extends Jsr250Metadata {

	/**
	 * The method that needs to be invoked on the interceptor.
	 */
	private final Method aroundInvokeMethod;

	/**
	 * The target method on the singleton that is intercepted.
	 */
	private final Method matchingMethod;

	/**
	 * Map from lifecycle event enumeration to callback method defined on a JEE
	 * component class. If there is no such callback method there will be no
	 * entry in the map.
	 */
	private Map<LifecycleEvent, List<Method>> lifecycleEventListenerMethods = new HashMap<LifecycleEvent, List<Method>>();

	/**
	 * Is this a class interceptor?
	 */
	private boolean classInterceptor;

	/**
	 * Is this a default interceptor?
	 */
	private boolean defaultInterceptor;

	/**
	 * Construct new method level interceptor metadata
	 * 
	 * @param name
	 * @param componentClass
	 * @param aroundInvokeMethod
	 * @param matchingMethod
	 * @param exclusion
	 */
	private InterceptorMetadata(String name, Class<?> componentClass, Method aroundInvokeMethod, Method matchingMethod) {
		// We don't put this level of detail--under component level--
		// into the root deployment unit
		super(new DeploymentUnitMetadata(), name, componentClass);
		this.aroundInvokeMethod = aroundInvokeMethod;
		this.matchingMethod = matchingMethod;
	}

	/**
	 * Construct new method level interceptor metadata with component name
	 * defaulting to the FQN of the component class
	 * 
	 * @param componentClass
	 * @param aroundInvokeMethod
	 * @param matchingMethod
	 */
	public InterceptorMetadata(Class<?> componentClass, Method aroundInvokeMethod, Method matchingMethod) {
		this(componentClass.getName(), componentClass, aroundInvokeMethod, matchingMethod);
	}

	/**
	 * Construct new class level interceptor metadata with component name
	 * defaulting to the FQN of the component class
	 * 
	 * @param componentClass
	 * @param aroundInvokeMethod
	 */
	public InterceptorMetadata(Class<?> componentClass, Method aroundInvokeMethod) {
		this(componentClass.getName(), componentClass, aroundInvokeMethod, null);
	}

	/**
	 * Create a new Interceptor metadata for a self interceptor. The component
	 * definition already exists.
	 * 
	 * @param jsr250
	 *            Jsr250 metadata we should wrap
	 * @param aroundInvokeMethod
	 *            aroundInvoke method to use on the instance
	 */
	public InterceptorMetadata(Jsr250Metadata jsr250, Method aroundInvokeMethod) {
		super(jsr250);
		this.aroundInvokeMethod = aroundInvokeMethod;
		this.matchingMethod = null;
	}

	public void registerLifecycleEventListenerMethod(LifecycleEvent le, Method m) {
		if (m != null) {
			if (!m.isAccessible()) {
				m.setAccessible(true);
			}
			List<Method> lifecycleListeners = lifecycleEventListenerMethods.get(le);
			if (lifecycleListeners == null)
					lifecycleListeners = new ArrayList<Method>();
			lifecycleListeners.add(m);
			lifecycleEventListenerMethods.put(le, lifecycleListeners);
			log.info("Interceptor callback listener: Registered " + m + " on " + getComponentClass().getName()
					+ " for " + le);
		}
	}

	/**
	 * Return the invocation-context style method for this callback
	 */
	public List<Method> getLifecycleEventListenerMethod(LifecycleEvent le) {
		return lifecycleEventListenerMethods.get(le);
	}

	public Method getAroundInvokeMethod() {
		return aroundInvokeMethod;
	}

	/**
	 * @return the method that this matches on, or null if it is a class level
	 *         interceptor
	 */
	public Method getMatchingMethod() {
		return this.matchingMethod;
	}

	/**
	 * @return the classInterceptor
	 */
	public boolean isClassInterceptor() {
		return classInterceptor;
	}

	/**
	 * @param classInterceptor
	 *            the classInterceptor to set
	 */
	public void setClassInterceptor(boolean classInterceptor) {
		this.classInterceptor = classInterceptor;
	}

	/**
	 * @return the defaultInterceptor
	 */
	public boolean isDefaultInterceptor() {
		return defaultInterceptor;
	}

	/**
	 * @param defaultInterceptor
	 *            the defaultInterceptor to set
	 */
	public void setDefaultInterceptor(boolean defaultInterceptor) {
		this.defaultInterceptor = defaultInterceptor;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.inject.Jsr250Metadata#setComponentContext(org.springframework.context.ApplicationContext,
	 *      org.springframework.beans.factory.support.BeanDefinitionRegistry)
	 */
	@Override
	public void setComponentContext(ApplicationContext beanFactoryForInstance, BeanDefinitionRegistry bdr) {
		super.setComponentContext(beanFactoryForInstance, bdr);
		registerBeanDefinition(bdr);
	}

	private void registerBeanDefinition(BeanDefinitionRegistry bdr) {
		// TODO: check if definition clashes are possible
		// if (bdr.containsBeanDefinition(getComponentName()))
		// throw new UnsupportedOperationException("bean definition clashes not
		// handled!");
		bdr.registerBeanDefinition(getComponentName(), getBeanDefinition());
	}


}
