/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.lang.reflect.Method;

import javax.interceptor.InvocationContext;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.springframework.aop.framework.ReflectiveMethodInvocation;
import org.springframework.aop.support.AopUtils;

/**
 * Spring AOP interceptor wrapping an EJB3 style interceptor.
 * 
 * @author Rod Johnson
 * @author Costin Leau
 * 
 */
public class JeeInterceptorInterceptor implements MethodInterceptor {

	private static final String INVOCATION_CONTEXT_KEY = InvocationContext.class.getName();

	private final InterceptorMetadata im;

	private final Method aroundInvokeMethod;

	private final Object jeeInterceptorInstance;

	public JeeInterceptorInterceptor(InterceptorMetadata im, Object jeeInterceptorInstance) {
		this.im = im;
		this.aroundInvokeMethod = im.getAroundInvokeMethod();
		this.jeeInterceptorInstance = jeeInterceptorInstance;
	}
	
	public JeeInterceptorInterceptor(Object jeeInterceptorInstance, Method aroundInvokeMethod) {
		this.jeeInterceptorInstance = jeeInterceptorInstance;
		this.aroundInvokeMethod = aroundInvokeMethod;
		this.im = null;
	}

	public InterceptorMetadata getInterceptorMetadata() {
		return this.im;
	}

	public Object getJeeInterceptorInstance() {
		return jeeInterceptorInstance;
	}

	public Object invoke(MethodInvocation mi) throws Throwable {		
		InvocationContext ic = lazyCreateInvocationContext((ReflectiveMethodInvocation) mi);
		// Use Spring AOP's exception unwrapping
		if (aroundInvokeMethod != null)
			return AopUtils.invokeJoinpointUsingReflection(jeeInterceptorInstance, aroundInvokeMethod, new Object[] { ic });
		else
			return jeeInterceptorInstance;
	}

	@SuppressWarnings("unchecked")
	private InvocationContext lazyCreateInvocationContext(ReflectiveMethodInvocation mi) {
		// No need to synchronize this, as the invocation chain will be processed in a single thread
		InvocationContext ic = (InvocationContext) mi.getUserAttributes().get(INVOCATION_CONTEXT_KEY);
		if (ic == null) {
			ic = new MethodInvocationInvocationContext(mi);
			mi.getUserAttributes().put(INVOCATION_CONTEXT_KEY, ic);
		}
		return ic;
	}

	@Override
	public String toString() {
		return "JeeInterceptorInterceptor: Wrapping <" + this.jeeInterceptorInstance + ">";
	}

}
