/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.lang.reflect.Method;
import java.util.Map;
import java.util.Set;

import org.springframework.aop.support.StaticMethodMatcherPointcutAdvisor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.intercept.InterceptionMetadata.InterceptorExclusion;

/**
 * Spring AOP pointcut that can be used for a method-targeted EJB3 style
 * interception
 * 
 * @author Rod Johnson
 */
public class JeeInterceptorPointcutAdvisor extends StaticMethodMatcherPointcutAdvisor {

	private final Method targetMethod;

	private final DeploymentUnitMetadata deploymentUnitMetadata;

	private final Map<Method, InterceptorExclusion> exclusions;

	private final Set<Object> interceptorOrderObject;

	/**
	 * 
	 * @param targetMethod
	 *            matching method. If this is null, will always match
	 * @param advice
	 *            JeeInterceptor
	 */
	public JeeInterceptorPointcutAdvisor(Method targetMethod, JeeInterceptorInterceptor advice,
			DeploymentUnitMetadata deploymentUnitMetadata, Map<Method, InterceptorExclusion> exclusion,
			Set<Object> interceptorOrderObject) {
		super(advice);
		this.targetMethod = targetMethod;
		this.deploymentUnitMetadata = deploymentUnitMetadata;
		this.exclusions = exclusion;
		this.interceptorOrderObject = interceptorOrderObject;
	}

	public boolean matches(Method method, Class targetClass) {
		// Note: As Spring AOP considers this a "static" pointcut, which
		// can be evaluated once only on proxy creation, there is no need
		// to optimize the evaluation here, as the result will be cached
		if (deploymentUnitMetadata.getContainerControlInterfaces().contains(method.getDeclaringClass())) {
			return false;
		}

		boolean result = !isExcluded(method);
		result &= !isInterceptorOrder(method, targetClass);
		if (this.targetMethod != null) {
			// match the method name
			result &= targetMethod.equals(method);
			// 
		}

		return result;
	}

	// isExplicitOrder
	private boolean isInterceptorOrder(Method method, Class targetClass) {
		return interceptorOrderObject.contains(method) || interceptorOrderObject.contains(targetClass);
	}

	protected boolean isExcluded(Method method) {
		InterceptorExclusion exclusion = exclusions.get(method);
		if (exclusion != null)
			return isClassExcluded(exclusion) || isDefaultExcluded(exclusion);
		return false;
	}

	private boolean isClassExcluded(InterceptorExclusion exclusion) {
		return (getAdvice().getInterceptorMetadata() != null && exclusion.isExcludeClassInterceptors() && getAdvice().getInterceptorMetadata().isClassInterceptor());
	}

	private boolean isDefaultExcluded(InterceptorExclusion exclusion) {
		return (getAdvice().getInterceptorMetadata() != null && exclusion.isExcludeDefaultInterceptors() && getAdvice().getInterceptorMetadata().isDefaultInterceptor());
	}

	/**
	 * Return the typed advice associated with this interceptor.
	 */
	public JeeInterceptorInterceptor getAdvice() {
		return (JeeInterceptorInterceptor) super.getAdvice();
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + ": matches=" + targetMethod + "; advice=" + getAdvice();
	}

}
