/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.util.List;

import org.springframework.jee.inject.LifecycleEvent;


/**
 * Interface introduced into JEE component proxies that allows
 * invocation of lifecycle interfaces.
 * <p>
 * Each singleton class may have a chain of interceptors. These may
 * have annotated (or otherwise configured) lifecycle listener methods,
 * distinct from the interceptor's own PostConstruct and other
 * lifecycle interfaces. As these methods are identified by annotations
 * or other metadata, there is no interface that can be used to 
 * drive them. This interface serves that purpose, and can be invoked
 * to invoke any callback event listener methods, in the order
 * of the interceptor chain.
 * <br>
 * See EJB 3.0 specification, 11.4.
 * <p>
 * Intended for container use only, hence __ prefix to name.
 * @author Rod Johnson
 */
public interface __ProxyControl {
	
	/**
	 * Invoke the given lifecycle method on the singleton instance
	 * and any interceptors that listen to it
	 * @param le lifefycle event
	 * @see LifecycleEvent
	 */
	void invokeLifecycleMethod(LifecycleEvent le);
	
	/**
	 * Return the metadata associated with this proxy
	 * @return the metadata associated with this proxy
	 */
	InterceptionMetadata getMetadata();

	/**
	 * Return the Spring AOP advisors wrapping JEE interceptors.
	 * Note that casting the proxy to the Spring AOP Advised proxy control interface
	 * will allow access to <b>all</b> Spring AOP advisors, including advisors
	 * wrapping JEE interceptors.
	 * @return advisors, from which we can obtain the JEE interceptor instances for
	 * the current proxy
	 */
	List<JeeInterceptorPointcutAdvisor> getJeeInterceptorAdvisors();

}
