/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.server;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.beans.factory.deployment.ContainerBootstrap;
import org.springframework.beans.factory.deployment.DefaultContainerBootstrap;
import org.springframework.beans.factory.xml.XmlBeanDefinitionReader;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.core.io.ResourceLoader;
import org.springframework.jee.inject.AnnotationComponentContributor;
import org.springframework.jee.inject.ComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Enricher;
import org.springframework.jee.validation.InvalidDeploymentUnitException;
import org.springframework.jee.validation.ValidationService;

/**
 * Bootstrap for the JEE container. Each JEE deployment is modelled as a child
 * of a server-level context, and each component also gets its own further
 * child. The deployment-level context contains a Spring singleton definition
 * for each EJB or other JEE component.
 * <p>
 * Uses the Spring IO abstraction for specifying resource location and resource
 * loading
 * 
 * @author Rod Johnson
 * @author Michael Chen
 * @see org.springframework.core.io.ResourceLoader
 */
public class Bootstrap {

	/**
	 * Location of server bootstrap definitions.
	 */
	private static final String SERVER_BOOTSTRAP_XML_LOCATION = "/org/springframework/jee/config/server/bootstrap.xml";

	/**
	 * Spring singleton definitions that provide starting point for each EJB
	 * deployment made by the server.
	 */
	private static final String DEPLOYMENT_BASE_XML_LOCATION = "/org/springframework/jee/config/deployment/deployment.xml";

	protected final Log log = LogFactory.getLog(getClass());

	/**
	 * Base application context representing server.
	 */
	private final GenericApplicationContext serverApplicationContext;

	private ContainerBootstrap containerBootstrap;

	private ValidationService validationService;

	/**
	 * Start the server-level context. This is capable of handling multiple JEE
	 * deployments.
	 */
	public Bootstrap(ContainerBootstrap containerBootstrap) {
		// TODO use bootstrap to start parent at server-level
		// Take args for locations

		this.containerBootstrap = containerBootstrap;
		serverApplicationContext = new GenericApplicationContext();

		XmlBeanDefinitionReader serverBdr = new XmlBeanDefinitionReader(serverApplicationContext);
		serverBdr.loadBeanDefinitions(SERVER_BOOTSTRAP_XML_LOCATION);
		serverApplicationContext.refresh();

		// Autowire this object from the server deployment, hitting its setters
		serverApplicationContext.getDefaultListableBeanFactory().autowireBeanProperties(this,
				AutowireCapableBeanFactory.AUTOWIRE_BY_TYPE, false);
	}

	/**
	 * Deploy the contributed components, ignoring standard or proprietary
	 * deployment descriptors.
	 */
	public Bootstrap() {
		this(new DefaultContainerBootstrap());
	}

	/**
	 * @param validationService the validationService to set
	 */
	public void setValidationService(ValidationService validationService) {
		this.validationService = validationService;
	}

	/**
	 * Return the application context associated with the server. This is the
	 * parent of the application contexts to which beans are deployed.
	 * 
	 * @return Server-level application context.
	 */
	public GenericApplicationContext getServerApplicationContext() {
		return this.serverApplicationContext;
	}

	/**
	 * Deploy the contributed components, ignoring standard or proprietary
	 * deployment descriptors.
	 * 
	 * @return an application context representing the deployment. This will be
	 *         a child of the server application context.
	 * @throws InvalidDeploymentUnitException
	 */
	public GenericApplicationContext deploy() throws InvalidDeploymentUnitException {
		return deploy(new AnnotationComponentContributor());
	}

	/**
	 * Deploy the contributed components, ignoring standard or proprietary
	 * deployment descriptors.
	 * 
	 * @param componentContributor source of component metadata
	 * @return an application context representing the deployment. This will be
	 *         a child of the server application context.
	 * @throws InvalidDeploymentUnitException
	 */
	public GenericApplicationContext deploy(ComponentContributor componentContributor)
			throws InvalidDeploymentUnitException {
		return deploy(null, componentContributor);
	}

	/**
	 * Deploy the contributed components, also looking for well-known resources
	 * using the given ResourceLoader.
	 * 
	 * @param resourceLoader ResourceLoader used to load resources (for example,
	 *            from a JAR)
	 * @param componentContributor source of component metadata
	 * @return an application context representing the deployment. This will be
	 *         a child of the server application context.
	 * @throws InvalidDeploymentUnitException
	 */
	public GenericApplicationContext deploy(ResourceLoader resourceLoader, ComponentContributor componentContributor)
			throws InvalidDeploymentUnitException {
		return deploy(new String[0], resourceLoader, componentContributor, null);
	}

	/**
	 * Deploy the contributed components, also looking for well-known resources
	 * using the given ResourceLoader.
	 * 
	 * @param requiredSpringLocations location of pure Spring files. Spring
	 *            resource paths.
	 * @param resourceLoader ResourceLoader used to load resources (for example,
	 *            from a JAR)
	 * @param componentContributor source of component metadata
	 * @param deploymentUnitMetadata may be null, in which case default metadata
	 *            will be used
	 * @return an application context representing the deployment. This will be
	 *         a child of the server application context.
	 * @throws InvalidDeploymentUnitException
	 */
	public GenericApplicationContext deploy(String[] requiredSpringLocations, ResourceLoader resourceLoader,
			ComponentContributor componentContributor, DeploymentUnitMetadata deploymentUnitMetadata)
			throws InvalidDeploymentUnitException {
		return deploy(requiredSpringLocations, resourceLoader, componentContributor, deploymentUnitMetadata,
				new String[0]);
	}

	/**
	 * Deploy using the given ComponentContributor, also looking for well-known
	 * resources using the given ResourceLoader. Enrichers will be used,
	 * normally included the standard deployment descriptor enhancer. Also load
	 * Spring XML files if specified.
	 * <p>
	 * Algorithm: first load annotations for classes, then deployment
	 * descriptors (standard, proprietary and Spring), the Spring XML files.
	 * 
	 * @param requiredSpringLocations location of pure Spring files. Spring
	 *            resource paths.
	 * @param optionalSpringLocations location of optional pure Spring files.
	 * @param resourceLoader ResourceLoader to use to load resources (for
	 *            example, from JAR). Standard and proprietary DDs will be
	 *            sought at this location
	 * @param componentContributor source of component metadata
	 * @param deploymentUnitMetadata DeploymentUnitMetadata for use. May be
	 *            null, in which case a default will be used.
	 * @return a GenericApplicationContext which will contain MetadataFactory
	 *         for the deployment and a Spring singleton definition for each EJB
	 *         (with the EJB name) and for each Spring singleton.
	 * @throws InvalidDeploymentUnitException if there are errors in the
	 *             metadata
	 */
	public GenericApplicationContext deploy(String[] requiredSpringLocations, ResourceLoader resourceLoader,
			ComponentContributor componentContributor, DeploymentUnitMetadata deploymentUnitMetadata,
			String... optionalSpringLocations) throws InvalidDeploymentUnitException {
		String[] requiredLocations = new String[requiredSpringLocations.length + 1];
		requiredLocations[0] = DEPLOYMENT_BASE_XML_LOCATION;
		System.arraycopy(requiredSpringLocations, 0, requiredLocations, 1, requiredSpringLocations.length);

		GenericApplicationContext deploymentContext = containerBootstrap.bootstrap(serverApplicationContext,
				requiredLocations, optionalSpringLocations, resourceLoader);

		Enricher enricher = new Enricher(deploymentContext);
		enricher.setValidationService(this.validationService);
		if (deploymentUnitMetadata != null) {
			enricher.setDeploymentUnitMetadata(deploymentUnitMetadata);
		}
		componentContributor.contribute(enricher);
		enricher.startup();
		return deploymentContext;
	}

}
