/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.transaction;

import java.rmi.RemoteException;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Abstract implementation containing the ejb specification shortcut rules:
 * <ul>
 * <li>if the throwable is Error or RemoteException then it is SystemException</li>
 * <li>if the throwable is CreateException/FinderException/RemoveException then it is an ApplicationException</li>
 * </ul>.
 * This class is meant only for internal usage and should be subclassed to add the rest of the 
 * cases and advanced functionality like exception caching. 
 * 
 * @author Costin Leau
 *
 */
public abstract class AbstractApplicationExceptionManager implements ApplicationExceptionManager {

	protected final Log log = LogFactory.getLog(getClass());
	
	/* (non-Javadoc)
	 * @see org.springframework.jee.transaction.ApplicationExceptionManager#isApplicationException(java.lang.Throwable)
	 */
	public boolean causesRollback(Throwable throwable) {
		
		// known system exceptions
		if (throwable instanceof Error || throwable instanceof RemoteException)
			return true;

		// known app exceptions
		if (throwable instanceof CreateException || throwable instanceof FinderException || throwable instanceof RemoveException)
			return false;
		
		return shouldCauseRollback((Exception) throwable);
	}

	/**
	 * Subclasses should extend this for more complex functionality.
	 * 
	 * @param exception
	 * @return
	 */
	protected abstract boolean shouldCauseRollback(Exception exception);
}
