/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.transaction;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.ejb.ApplicationException;

import org.springframework.jee.inject.ApplicationExceptionInfo;

/**
 * Default implementation for detecting application exceptions.
 * 
 * @author Costin Leau
 * 
 */
public class DefaultApplicationExceptionManager extends AbstractApplicationExceptionManager {
	// TODO: do some benchmark and see if this algorithm is good enough
	// TODO: what about jdk 5.0 ConcurrentHashMap?

	private Map<Class<? extends Exception>, ApplicationExceptionInfo> checkedApplicationExceptions = Collections.synchronizedMap(new HashMap<Class<? extends Exception>, ApplicationExceptionInfo>());

	private Map<Class<? extends RuntimeException>, ApplicationExceptionInfo> uncheckedApplicationExceptions = Collections.synchronizedMap(new HashMap<Class<? extends RuntimeException>, ApplicationExceptionInfo>());

	// TODO: can we actually have checked System exceptions?
	private Set<Class<? extends Exception>> checkedSystemExceptions = Collections.synchronizedSet(new HashSet<Class<? extends Exception>>());

	private Set<Class<? extends RuntimeException>> uncheckedSystemExceptions = Collections.synchronizedSet(new HashSet<Class<? extends RuntimeException>>());

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.transaction.AbstractApplicationExceptionManager#analyzeThrowable(java.lang.Throwable)
	 */
	@Override
	protected boolean shouldCauseRollback(Exception exception) {
		// 1. check exception against the registered app exceptions (we assume
		// that
		// in general there will be a lot more system exceptions then
		// application exceptions)

		ApplicationExceptionInfo info;
		Class exceptionClass = exception.getClass();
		boolean isRuntimeException = exception instanceof RuntimeException;

		if (isRuntimeException) {
			info = uncheckedApplicationExceptions.get(exceptionClass);
		}
		else
			info = checkedApplicationExceptions.get(exceptionClass);

		if (info != null) {
			return info.isRollback();
		}

		// 2. if not found check it against the system exceptions
		boolean systemException = false;

		if (isRuntimeException)
			systemException = uncheckedSystemExceptions.contains(exceptionClass);
		else
			systemException = checkedSystemExceptions.contains(exceptionClass);

		if (systemException)
			return true;

		// if still not found analyze it and place inside the cache
		if (isRuntimeException) {
			Class<? extends RuntimeException> runtimeExceptionClass = (Class<? extends RuntimeException>) exceptionClass;
			ApplicationExceptionInfo appInfo = inspectRuntimeException(runtimeExceptionClass);
			// system exception
			if (appInfo == null) {
				uncheckedSystemExceptions.add(runtimeExceptionClass);
				return true;
			}
			// application exception
			else {
				uncheckedApplicationExceptions.put(runtimeExceptionClass, appInfo);
				return appInfo.isRollback();
			}
		}

		// only application exceptions can be checked
		ApplicationExceptionInfo exInfo = createApplicationExceptionInfo(exceptionClass);
		checkedApplicationExceptions.put((Class<? extends Exception>) exceptionClass, exInfo);
		return exInfo.isRollback();
	}

	/**
	 * Adds a list of application exceptions discovered through introspection
	 * (only checked exceptions are considered).
	 * 
	 * @param appExceptions
	 */
	public void addCheckedApplicationExceptions(List<Class<? extends Exception>> appExceptions) {
		for (Class<? extends Exception> appException : appExceptions)
			checkedApplicationExceptions.put(appException, createApplicationExceptionInfo(appException));
	}

	/**
	 * Create AppExInfo for checked application exceptions.
	 * 
	 * @param appException
	 * @return
	 */
	private ApplicationExceptionInfo createApplicationExceptionInfo(Class<?> appException) {
		// TODO: is this according to the spec ?
		return ApplicationExceptionInfo.COMMIT;
	}

	/**
	 * Create AppExInfo for unchecked application exceptions.
	 * 
	 * @param uncheckedException
	 * @return
	 */
	protected ApplicationExceptionInfo inspectRuntimeException(Class<? extends RuntimeException> uncheckedException) {
		ApplicationException annotation = uncheckedException.getAnnotation(ApplicationException.class);
		if (annotation == null)
			return null;
		return (annotation.rollback() ? ApplicationExceptionInfo.ROLLBACK : ApplicationExceptionInfo.COMMIT);

	}

	/**
	 * Add a map of application exceptions that are already parsed (for example
	 * defined within an Ejb container through an xml file). The map can contain
	 * both checked and unchecked exceptions.
	 * 
	 * @param applicationExceptions
	 */
	@SuppressWarnings("unchecked")
	public void addApplicationException(Map<Class<? extends Exception>, ApplicationExceptionInfo> applicationExceptions) {

		for (Map.Entry<Class<? extends Exception>, ApplicationExceptionInfo> entry : applicationExceptions.entrySet()) {
			if (RuntimeException.class.isAssignableFrom(entry.getKey())) {
				uncheckedApplicationExceptions.put((Class<? extends RuntimeException>) entry.getKey(), entry.getValue());
			}
			else
				checkedApplicationExceptions.put(entry.getKey(), entry.getValue());
		}
	}
}
