/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.transaction;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.ejb.TransactionAttribute;

import org.springframework.aop.framework.ProxyFactory;
import org.springframework.jee.config.ServiceEnvironment;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.intercept.InterceptionMetadata;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.interceptor.DefaultTransactionAttribute;
import org.springframework.transaction.interceptor.NameMatchTransactionAttributeSource;
import org.springframework.transaction.interceptor.TransactionAttributeSourceAdvisor;
import org.springframework.transaction.interceptor.TransactionInterceptor;

/**
 * Metadata for a transactional component such as an EJB. Adds transaction
 * management to DI and interception capability.
 * 
 * @author Rod Johnson
 * @author Costin Leau
 * 
 */
public class TransactionalMetadata extends InterceptionMetadata {

	private Map<Method, TransactionAttribute> transactionalMethodAttributes = new HashMap<Method, TransactionAttribute>();

	private List<Class<? extends Exception>> applicationCheckedExceptions = new ArrayList<Class<? extends Exception>>();

	/**
	 * Short-cut flag which marks if the given component is transactional or
	 * not.
	 */
	private boolean transactional = false;

	/**
	 * @param name
	 * @param componentClass
	 */
	public TransactionalMetadata(DeploymentUnitMetadata dum, String name, Class<?> componentClass) {
		super(dum, name, componentClass);
	}

	/*
	 * Add the tx interceptors (called from superclass when creating the
	 * interception proxy).
	 * 
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.jee.intercept.InterceptionMetadata#addComponentContractInterceptors(org.springframework.aop.framework.ProxyFactory)
	 */
	@Override
	protected void addComponentContractInterceptors(ProxyFactory pf) {

		// TODO: should this throw an exception instead?
		if (!isTransactional() || getTransactionManager() == null)
			return;

		// TODO: these can be a singleton - are there any side-effects if they
		// are shared?
		DefaultApplicationExceptionManager appExceptionManager = new DefaultApplicationExceptionManager();
		//	add metadata appExceptions
		appExceptionManager.addCheckedApplicationExceptions(applicationCheckedExceptions);

		// add deployment unit appExceptions
		appExceptionManager.addApplicationException(getDeploymentUnitMetadata().getApplicationExceptions());

		NameMatchTransactionAttributeSource transactionAttributeSource = new NameMatchTransactionAttributeSource();
		// add tx attributes
		Map<String, org.springframework.transaction.interceptor.TransactionAttribute> txAttributes = new HashMap<String, org.springframework.transaction.interceptor.TransactionAttribute>(
				transactionalMethodAttributes.size());
		for (Map.Entry<Method, TransactionAttribute> methodAttribute : transactionalMethodAttributes.entrySet()) {
			txAttributes.put(methodAttribute.getKey().getName(), copyEjbTxAttribute(methodAttribute.getValue(),
					new Ejb3TransactionAttribute(appExceptionManager)));
		}

		transactionAttributeSource.setNameMap(txAttributes);
		TransactionInterceptor txInterceptor = new TransactionInterceptor(getTransactionManager(),
				transactionAttributeSource);

		// Rely on default pointcut.
		pf.addAdvisor(new TransactionAttributeSourceAdvisor(txInterceptor));
	}

	/**
	 * Return the transaction manager to be used for the tx support.
	 * 
	 * @return
	 */
	private PlatformTransactionManager getTransactionManager() {
		ServiceEnvironment env = this.getDeploymentUnitMetadata().getEnvironment();
		return (env == null ? null : env.getTransactionManager());
	}

	/**
	 * Sets the given Spring transaction definition according to the passed in
	 * ejb3 attribute. Required for the null case of annotation.
	 * 
	 * @param ejbTxAttribute
	 * @param springTxAttribute
	 */
	protected DefaultTransactionAttribute copyEjbTxAttribute(TransactionAttribute ejbTxAttribute,
			DefaultTransactionAttribute springTxAttribute) {
		if (ejbTxAttribute != null)
			springTxAttribute.setPropagationBehaviorName("PROPAGATION_" + ejbTxAttribute.value().name());
		return springTxAttribute;
	}

	/**
	 * Add transactional method metadata.
	 * 
	 * @param methodOnClass
	 * @param txAttribute
	 */
	public void addTransactionalMethod(Method methodOnClass, TransactionAttribute txAttribute) {
		// each method can have only one attribute (no need to check for
		// clashes)
		transactionalMethodAttributes.put(methodOnClass, txAttribute);
	}

	/**
	 * Add application exception metadata.
	 * 
	 * @param exception
	 */
	public void addApplicationCheckedException(Class<? extends Exception> exception) {
		applicationCheckedExceptions.add(exception);
	}

	protected List<Class<? extends Exception>> getApplicationCheckedExceptions() {
		return applicationCheckedExceptions;
	}

	protected Map<Method, TransactionAttribute> getTransactionalMethodAttributes() {
		return transactionalMethodAttributes;
	}

	/**
	 * @return Returns the transactional.
	 */
	public boolean isTransactional() {
		return transactional;
	}

	/**
	 * @param transactional
	 *            The transactional to set.
	 */
	public void setTransactional(boolean transactional) {
		this.transactional = transactional;
	}

}
