/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.validation;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.util.ReflectionUtils;

/**
 * Implementation of ValidationService that reflectively invokes annotated rule
 * methods.
 * <p>
 * The signature of a valid validator method looks as follows:
 * <code>
 * @EJB_3_0(chapter = 3, heading = 2)
	public void checkHasBusinessInterfaces(InterceptionMetadata im) throws SpecificationViolation {
 * </code>
 * 
 * @author Rod Johnson
 * 
 */
public class ReflectiveRuleBasedValidationService implements ValidationService {

	private Log log = LogFactory.getLog(getClass());

	/**
	 * Map from validation method to validator instance
	 */
	private List<SpecificationRule> rules = new LinkedList<SpecificationRule>();
	
	private Class[] allAnnotations = findAllPlatformValidationRuleAnnotations();
	
	private static Class[] findAllPlatformValidationRuleAnnotations() {
		Set<Class> annotations = new HashSet<Class>();
		for (Platform p : Platform.values()) {
			for (Class annotationClass : p.getAnnotations()) {
				annotations.add(annotationClass);
			}
		}
		Class[] allAnnotations = new Class[annotations.size()];
		int i = 0;
		for (Class annotationClass : annotations) {
			allAnnotations[i++] = annotationClass;
		}
		return allAnnotations;
	}

	/**
	 * Add an instance of a class that may be a validator,
	 * caching validation methods.
	 * @param instance instance of a class that may be a validator.
	 * It is not an error if the class contains no validation methods.
	 * @return the number of validation methods found
	 */
	public int addPotentialValidator(final Object instance) {
		int oldCount = rules.size();
		
		if (!instance.getClass().isAnnotationPresent(SpecificationRules.class)) {
			log.info("No specification rules in " + instance.getClass());
		}
		
		ReflectionUtils.doWithMethods(instance.getClass(), new ReflectionUtils.MethodCallback() {
			public void doWith(Method candidateValidationMethod) throws IllegalArgumentException, IllegalAccessException {
				List<SpecificationRule> ruleAnnotations = findRuleAnnotations((Class<? extends Annotation>[])allAnnotations, instance, candidateValidationMethod);
				for (SpecificationRule specRule : ruleAnnotations) {
					rules.add(specRule);
				}
			}
		}, new ReflectionUtils.MethodFilter() {
			public boolean matches(Method candidateValidationMethod) {
				return 	candidateValidationMethod.getParameterTypes().length == 1 &&
						Jsr250Metadata.class.isAssignableFrom(candidateValidationMethod.getParameterTypes()[0]) &&
						candidateValidationMethod.getExceptionTypes().length == 1 &&
						candidateValidationMethod.getExceptionTypes()[0].equals(SpecificationViolation.class);
			}
		});
		int rulesFound = rules.size() - oldCount;
		log.info(rulesFound + " specification rules found in " + instance.getClass());
		return rulesFound;
	}

	/**
	 * Validate the given deployment
	 * @param dum metadata for deployment unit to validate 
	 * @return the number of components validated
	 */
	public int validate(DeploymentUnitMetadata dum) throws InvalidDeploymentUnitException {
		InvalidDeploymentUnitException idum = new InvalidDeploymentUnitException();
		int componentsValidated = 0;
		for (Jsr250Metadata component : dum.getDeployedComponentMetadata()) {
			invokeValidationMethods(component, idum);
			++componentsValidated;
		}

		if (idum.hasErrors()) {
			throw idum;
		}
		return componentsValidated;
	}

	/**
	 * Reflectively invoke all the validation methods on the given component
	 * @param component component we have metadata for
	 * @param idum exception for whole deployment unit, allowing us to
	 * add further errors
	 */
	protected void invokeValidationMethods(final Jsr250Metadata component, final InvalidDeploymentUnitException idum) {

		for (SpecificationRule location : rules) {			
			if (location.getPlatforms().contains(component.getDeploymentUnitMetadata().getTargetPlatform()) &&
					location.appliesTo(component.getClass())) {
				location.execute(component, idum);
			}
		}
	}

	private List<SpecificationRule> findRuleAnnotations(Class<? extends Annotation>[] annotations, Object instance, 
			Method candidateValidationMethod) {
		List<SpecificationRule> locations = new LinkedList<SpecificationRule>();
		for (Class<? extends Annotation> annotationClass : annotations) {
			Object annotation = JeeAnnotationUtils
				.findMethodAnnotation(annotationClass, candidateValidationMethod, instance.getClass());
			if (annotation != null) {
				locations.add(new AnnotatedMethodSpecificationRule(annotationClass, annotation, candidateValidationMethod, instance));
			}
		}
		return locations;
	}

}
