package org.springframework.jee;

import java.lang.reflect.Method;
import java.util.List;

import javax.annotation.Resource;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;

import junit.framework.TestCase;

/**
 * Test case for the jsr250 inheritance mechanism using the example taken from
 * the specs.
 * 
 * @author Costin Leau
 * 
 */
public class JeeAnnotationInheritanceTests extends TestCase {

	@TransactionAttribute(TransactionAttributeType.REQUIRED)
	class Base {
		@Resource(name="Base")
		protected Object foo;

		@TransactionAttribute(TransactionAttributeType.NOT_SUPPORTED)
		public void foo() {
		}

		public void bar() {
		}
	}

	class DerivedA extends Base {

		public void foo() {
		}
	}

	public class DerivedB extends Base {
		@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
		public void foo() {
		}
	}

	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	@Resource(name="DerivedC")
	public class DerivedC extends Base {
		public Object foo;
		public Object bar;

		public void foo() {
		}

		public void bar() {
		}
	}

	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	public class DerivedD extends Base {
		public void bar() {
		}
	}

	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	@Resource
	public class DerivedE extends Base {
		public Object foo;
		@Resource(name="DerivedE")
		public Object bar;
	}

	public void testTransactionInheritance() {
		// foo method
		String fooMethodName = "foo";
		String barMethodName = "bar";

		TransactionAttribute fooDerivedA = JeeAnnotationUtils.findMethodAnnotation(TransactionAttribute.class,
				DerivedA.class, fooMethodName);
		TransactionAttribute fooDerivedB = JeeAnnotationUtils.findMethodAnnotation(TransactionAttribute.class,
				DerivedB.class, fooMethodName);
		TransactionAttribute fooDerivedC = JeeAnnotationUtils.findMethodAnnotation(TransactionAttribute.class,
				DerivedC.class, fooMethodName);
		TransactionAttribute barDerivedD = JeeAnnotationUtils.findMethodAnnotation(TransactionAttribute.class,
				DerivedD.class, barMethodName);
		TransactionAttribute barDerivedE = JeeAnnotationUtils.findMethodAnnotation(TransactionAttribute.class,
				DerivedE.class, barMethodName);

		// see why annotation is null

		assertNull(fooDerivedA);
		assertTrue(testTransactionAttributeValue(fooDerivedB, TransactionAttributeType.REQUIRES_NEW));
		assertTrue(testTransactionAttributeValue(fooDerivedC, TransactionAttributeType.REQUIRES_NEW));
		assertTrue(testTransactionAttributeValue(barDerivedD, TransactionAttributeType.REQUIRES_NEW));

		assertNull(barDerivedE);
		// below is the test that confirms the jsr250 examples which I think are
		// broken
		// assertTrue(testTransactionAttributeValue(barDerivedE,
		// TransactionAttributeType.REQUIRED));
	}

	protected boolean testTransactionAttributeValue(TransactionAttribute annotation, TransactionAttributeType value) {
		if (annotation == null)
			throw new IllegalArgumentException("annotation is required");
		if (annotation.value() == null && value == null)
			return true;

		return annotation.value().equals(value);
	}

	public void testfindFieldAnnotation() throws Exception {
		String foo = "foo";
		String bar = "bar";
		Class fieldType = Object.class;

		Resource fooDerivedA = JeeAnnotationUtils.findFieldAnnotation(Resource.class, DerivedA.class, foo, fieldType);
		Resource fooDerivedB = JeeAnnotationUtils.findFieldAnnotation(Resource.class, DerivedB.class, foo, fieldType);
		Resource fooDerivedC = JeeAnnotationUtils.findFieldAnnotation(Resource.class, DerivedC.class, foo, fieldType);
		Resource barDerivedD = JeeAnnotationUtils.findFieldAnnotation(Resource.class, DerivedD.class, bar, fieldType);
		Resource barDerivedE = JeeAnnotationUtils.findFieldAnnotation(Resource.class, DerivedE.class, bar, fieldType);
		
		assertEquals("Base", fooDerivedA.name());
		assertEquals("Base", fooDerivedB.name());
		assertNull(fooDerivedC);
		assertNull(barDerivedD);
		assertNotNull("DerivedE", barDerivedE.name());
	}
	
	public void testFindAnnotatedMethods() throws Exception
	{
		List<Method> methods = JeeAnnotationUtils.findAnnotatedMethods(Base.class, TransactionAttribute.class);
		assertEquals(1, methods.size());
		assertEquals("foo", methods.get(0).getName());
		methods = JeeAnnotationUtils.findAnnotatedMethods(DerivedC.class, TransactionAttribute.class);
		assertEquals(1, methods.size());
		assertEquals("foo", methods.get(0).getName());

	}
}
