/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.jee.ejb;

import java.util.List;

import javax.ejb.EJB;
import javax.ejb.Stateful;
import javax.ejb.Stateless;

import junit.framework.TestCase;

import org.springframework.aop.support.AopUtils;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.jee.inject.ChainedProcessorsComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Injection;
import org.springframework.jee.inject.MetadataProcessor;
import org.springframework.jee.server.Bootstrap;
import org.springframework.jee.server.EjbBootstrap;
import org.springframework.jee.validation.InvalidDeploymentUnitException;

import sample.EngineControl;
import sample.EngineControlBean;
import sample.LaundryMachine;

public class EjbTests extends TestCase {

	public List<MetadataProcessor> processors;

	GenericApplicationContext jarAc;

	@Stateless
	private class FieldEjbReference implements EjbLaundryMachine {
		// field setter
		@EJB
		private EngineControl control;

		public EngineControl getAnotherEngine() {
			return control;
		}
	}

	@Stateful
	private class MethodEjbReference implements EjbLaundryMachine {

		private EngineControl foo;

		public EngineControl getAnotherEngine() {
			return foo;
		}

		@EJB
		public void setFoo(EngineControl engineControl) {
			this.foo = engineControl;
		}
	}

	protected void setUp() throws Exception {
		ChainedProcessorsComponentContributor ms = new EjbAnnotationComponentContributor(AtEjbLaundryMachineBean.class,
				EngineControlBean.class);

		Bootstrap bootstrap = new EjbBootstrap();
		jarAc = bootstrap.deploy(ms);
	}

	protected void tearDown() throws Exception {
		jarAc.close();
		jarAc = null;
	}

	public void testMethodEjbReferenceInjection() throws InvalidDeploymentUnitException {

		LaundryMachine lm = (LaundryMachine) jarAc.getBean(AtEjbLaundryMachineBean.class.getSimpleName());
		assertFalse("Should not be proxied", AopUtils.isAopProxy(lm));
		assertNotNull("EJB was not injected using @EJB", lm.getEngineControl());
	}

	public void testFieldEjbReferenceInjection() throws Exception {

		EjbLaundryMachine lm = (EjbLaundryMachine) jarAc.getBean(AtEjbLaundryMachineBean.class.getSimpleName());
		assertFalse("Should not be proxied", AopUtils.isAopProxy(lm));
		EngineControl reference = lm.getAnotherEngine();
		assertNotNull("EJB was not injected using @EJB", reference);
	}

	public void testFieldInjectionsDefaultValues() throws Exception {
		Class clazz = FieldEjbReference.class;
		EjbInjectionMetadata metadata = new EjbInjectionMetadata(new DeploymentUnitMetadata(), clazz.getSimpleName(),
				clazz);

		EjbReferenceProcessor processor = new EjbReferenceProcessor();
		processor.addEjbReferenceMetadata(metadata);

		assertEquals("inadequate number of ejb injections", 1, metadata.getEjbInjections().size());

		Injection injection = metadata.getEjbInjections().get(0);
		assertEquals("Field name not computed correctly", FieldEjbReference.class.getName() + "/control",
				injection.getName());
		assertSame("Field type not computed correctly", EngineControl.class, injection.getType());
	}

	public void testMethodInjectionsDefaultValues() throws Exception {
		Class clazz = MethodEjbReference.class;
		EjbInjectionMetadata metadata = new EjbInjectionMetadata(new DeploymentUnitMetadata(), clazz.getSimpleName(),
				clazz);

		EjbReferenceProcessor processor = new EjbReferenceProcessor();
		processor.addEjbReferenceMetadata(metadata);

		assertEquals("inadequate number of ejb injections", 1, metadata.getEjbInjections().size());

		Injection injection = metadata.getEjbInjections().get(0);
		assertEquals("Method name not computed correctly", MethodEjbReference.class.getName() + "/foo",
				injection.getName());
		assertSame("Method type not computed correctly", EngineControl.class, injection.getType());
	}
}
