/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.inject;

import java.lang.reflect.Method;

import junit.framework.TestCase;

import org.springframework.jee.intercept.InterceptionMetadata;
import org.springframework.jee.intercept.InterceptorMetadata;

import sample.CountInterceptor;
import sample.LaundryMachineBean;
import sample.PostConstructBeanRelyingOnInterceptor;
import sample.PostConstructInterceptor;
import sample.intercept.InterceptedLaundryMachine;

/**
 * Unit tests for AnnotationComponentContributor
 * 
 * @author Rod Johnson
 * 
 */
public class AnnotationComponentContributorBaseClassTests extends TestCase {

	private DummyAnnotationComponentContributor acc = new DummyAnnotationComponentContributor();

	private class DummyAnnotationComponentContributor extends
			ChainedProcessorsComponentContributor{
		public void contribute(Enricher enricher) {
			throw new UnsupportedOperationException();
		}
	}

	public void testNoInterceptors() {
		String componentName = "foobar";
		InterceptionMetadata im = (InterceptionMetadata) acc.findComponentMetadata(componentName,
				LaundryMachineBean.class, createDeploymentUnitMetadata());
		assertEquals(componentName, im.getComponentName());
		assertEquals("No interceptors", 0, im.getInterceptorMetadata().size());
		assertEquals("One injection", 1, im.getInjections().size());
		Injection inj = im.getInjections().get(0);
		assertEquals("EngineControlBean", inj.getName());
		assertEquals("EngineControlBean", inj.getInfo().getName());
	}

	protected DeploymentUnitMetadata createDeploymentUnitMetadata() {
		//GenericApplicationContext deploymentContext = new GenericApplicationContext();
		return new DeploymentUnitMetadata();
	}

	public void testInterceptorsAndInheritedLifecycleMethods() {
		String componentName = "foobar";
		InterceptionMetadata im = (InterceptionMetadata) acc.findComponentMetadata(componentName,
				InterceptedLaundryMachine.class, createDeploymentUnitMetadata());
		assertEquals(componentName, im.getComponentName());
		assertEquals("Two interceptors", 2, im.getInterceptorMetadata().size());
		assertEquals("One injection", 1, im.getInjections().size());
		Injection inj = im.getInjections().get(0);
		assertEquals("EngineControlBean", inj.getName());
		assertEquals("EngineControlBean", inj.getInfo().getName());

		// Test inherited childPreDestroy method
		Method preDestroyMethod = im
				.getLifecycleEventCallbackMethod(LifecycleEvent.PRE_DESTROY).get(0);
		assertNotNull("Found inherited childPreDestroy method", preDestroyMethod);
		assertEquals("nameDoesntMatter", preDestroyMethod.getName());

		assertNotNull("Noop childPostConstruct found on leaf", im
				.getLifecycleEventCallbackMethod(LifecycleEvent.POST_CONSTRUCT));
		assertNull("No prePassivate", im
				.getLifecycleEventCallbackMethod(LifecycleEvent.PRE_PASSIVATE));
		assertNull("No postActivate", im
				.getLifecycleEventCallbackMethod(LifecycleEvent.POST_ACTIVATE));

		// Examine the first interceptor
		InterceptorMetadata imd = im.getInterceptorMetadata().get(0);
		assertEquals(CountInterceptor.class, imd.getComponentClass());
		assertEquals(0, imd.getInjections().size());

		imd = im.getInterceptorMetadata().get(1);
		assertEquals(PostConstructInterceptor.class, imd.getComponentClass());
		assertEquals(0, imd.getInjections().size());
	}

	public void testInterceptorsWithLifecycleMethods() {
		String componentName = "foobar";
		InterceptionMetadata im = (InterceptionMetadata) acc.findComponentMetadata(componentName,
				PostConstructBeanRelyingOnInterceptor.class, createDeploymentUnitMetadata());

		for (LifecycleEvent le : LifecycleEvent.values()) {
			assertNull(im.getLifecycleEventCallbackMethod(le));
		}

		// Examine the single interceptor
		assertEquals("One method interceptor", 1, im.getInterceptorMetadata()
				.size());
		InterceptorMetadata imd = im.getInterceptorMetadata().get(0);
		assertEquals(PostConstructInterceptor.class, imd.getComponentClass());
		assertEquals(0, imd.getInjections().size());

		// Check that listener methods were picked up
		assertNotNull(imd.getAroundInvokeMethod());

		Method postConstructForInterceptorItself = imd
				.getLifecycleEventCallbackMethod(LifecycleEvent.POST_CONSTRUCT).get(0);
		assertNotNull("Interceptor has post construct listener",
				postConstructForInterceptorItself);
		assertEquals("anyName", postConstructForInterceptorItself.getName());
		assertEquals(0,
				postConstructForInterceptorItself.getParameterTypes().length);
		assertNull("No prePassivate", imd
				.getLifecycleEventCallbackMethod(LifecycleEvent.PRE_PASSIVATE));
		assertNull("No postActivate", imd
				.getLifecycleEventCallbackMethod(LifecycleEvent.POST_ACTIVATE));
		assertNull("No childPreDestroy", imd
				.getLifecycleEventCallbackMethod(LifecycleEvent.PRE_DESTROY));

		Method postConstructListenerOnInterceptor = imd
				.getLifecycleEventListenerMethod(LifecycleEvent.POST_CONSTRUCT).get(0);
		assertNotNull("Interceptor has post construct listener also",
				postConstructListenerOnInterceptor);
		assertEquals("anyName", postConstructListenerOnInterceptor.getName());
		assertEquals("PostConstruct listener takes InvocationContext", 1,
				postConstructListenerOnInterceptor.getParameterTypes().length);
		assertNull("No prePassivate", imd
				.getLifecycleEventListenerMethod(LifecycleEvent.PRE_PASSIVATE));
		assertNull("No postActivate", imd
				.getLifecycleEventListenerMethod(LifecycleEvent.POST_ACTIVATE));
		assertNull("No childPreDestroy", imd
				.getLifecycleEventListenerMethod(LifecycleEvent.PRE_DESTROY));
	}
}
