package org.springframework.jee.intercept;

import java.lang.reflect.Method;

import junit.framework.TestCase;

import org.springframework.aop.support.AopUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.inject.ChainedProcessorsComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.InterceptionProcessor;
import org.springframework.jee.inject.Jsr250Processor;
import org.springframework.jee.inject.LifecycleEvent;
import org.springframework.jee.inject.MetadataProcessor;
import org.springframework.jee.server.Bootstrap;
import org.springframework.jee.server.TrackingAnnotationComponentContributor;
import org.springframework.jee.transaction.BusinessInterface;
import org.springframework.jee.validation.InvalidDeploymentUnitException;

import sample.EngineControlBean;
import sample.LaundryMachine;
import sample.LaundryMachineBean;

public class InheritanceLifecycleTests extends TestCase {

	private InterceptorMetadata baseInterceptorMetadata;

	protected void setUp() throws Exception {
		super.setUp();
		baseInterceptorMetadata = InterceptorInheritanceTests.createBaseInterceptorMetadata(ChildSelfInterceptBean.class);

	}

	protected void tearDown() throws Exception {
		super.tearDown();
		baseInterceptorMetadata = null;
	}

	public void testBeanLifecycle() throws Exception {
		Bootstrap bootstrap = new Bootstrap();
		TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(
				ChildSelfInterceptBean.class);
		// change the default processor stack (and use only interception)
		tcc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
		ChildSelfInterceptBean.childPostConstruct = false;
		ChildSelfInterceptBean.childPreDestroy = false;
		SelfInterceptBean.postConstruct = false;
		SelfInterceptBean.preDestroy = false;
		BaseSelfInterceptBean.basePreDestroy = false;
		BaseSelfInterceptBean.basePostConstruct = false;

		ApplicationContext jarAc = bootstrap.deploy(tcc);
		BusinessInterface bi = (BusinessInterface) jarAc.getBean(ChildSelfInterceptBean.class.getSimpleName());
		assertTrue("Should be proxied", AopUtils.isAopProxy(bi));
		// check bean
		assertTrue("PreConstruct was not called", ChildSelfInterceptBean.childPostConstruct);
		assertFalse("PreDestroy was called", ChildSelfInterceptBean.childPreDestroy);

		// check super classes
		assertTrue("PreConstruct was not called", SelfInterceptBean.postConstruct);
		assertFalse("PreDestroy was called", SelfInterceptBean.preDestroy);
		// check super classes
		assertTrue("PreConstruct was not called", BaseSelfInterceptBean.basePostConstruct);
		assertFalse("PreDestroy was called", BaseSelfInterceptBean.basePreDestroy);

		// Programmatically raise event
		tcc.getDeployedComponents().get(0).invokeLifecycleMethods(bi, LifecycleEvent.PRE_DESTROY);

		// test bean
		assertTrue("PreDestroy was not called", ChildSelfInterceptBean.childPreDestroy);

		// test super classes
		assertTrue("PreDestroy was not called", SelfInterceptBean.preDestroy);
		assertTrue("PreDestroy was not called", BaseSelfInterceptBean.basePreDestroy);

	}

	public void testDefaultInterceptorLifecycle() throws Exception {
		Bootstrap bootstrap = new Bootstrap();
		TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(
				ChildSelfInterceptBean.class, EngineControlBean.class);
		// change the default processor stack (and use only interception)
		tcc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
		DefaultNephewInterceptorWithDependencies.nephewPostConstruct = false;
		DefaultNephewInterceptorWithDependencies.nephewPreDestroy = false;
		DefaultChildInterceptorWithDependencies.childPreDestroy = false;
		DefaultChildInterceptorWithDependencies.childPostConstruct = false;

		DeploymentUnitMetadata unit = new DeploymentUnitMetadata();
		Class clazz = DefaultNephewInterceptorWithDependencies.class;
		Method aroundInvokeMethod = JeeAnnotationUtils.findAroundInvokeMethod(clazz);
		InterceptionProcessor processor = new InterceptionProcessor();
		
		InterceptorMetadata defaultInterceptor = processor.createInterceptorMetadata(clazz, aroundInvokeMethod);
		unit.registerDefaultInterceptorMetadata(defaultInterceptor);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, tcc, unit);

		BusinessInterface bi = (BusinessInterface) jarAc.getBean(ChildSelfInterceptBean.class.getSimpleName());
		assertTrue("Should be proxied", AopUtils.isAopProxy(bi));
		// check bean
		assertTrue("PreConstruct was not called", DefaultNephewInterceptorWithDependencies.nephewPostConstruct);
		assertFalse("PreDestroy was called", DefaultNephewInterceptorWithDependencies.nephewPreDestroy);

		// check super classes
		assertTrue("PreConstruct was not called", DefaultChildInterceptorWithDependencies.childPostConstruct);
		assertFalse("PreDestroy was called", DefaultChildInterceptorWithDependencies.childPreDestroy);

		// Programmatically raise event
		tcc.getDeployedComponents().get(0).invokeLifecycleMethods(bi, LifecycleEvent.PRE_DESTROY);

		// check base interceptor
		assertTrue("PreDestroy was not called", DefaultNephewInterceptorWithDependencies.nephewPreDestroy);

		// test super interceptor classes
		assertTrue("PreDestroy was not called", DefaultChildInterceptorWithDependencies.childPreDestroy);
	}

	public void testBeanLifecycleOrderAndShortcircuit() throws Exception {

		Bootstrap bootstrap = new Bootstrap();
		TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(
				ChildSelfInterceptBean.class, EngineControlBean.class);
		// change the default processor stack (and use only interception)
		tcc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
		DefaultNephewInterceptorWithDependencies.nephewPostConstruct = false;
		DefaultNephewInterceptorWithDependencies.nephewPreDestroy = false;
		DefaultChildInterceptorWithDependencies.childPreDestroy = false;
		DefaultChildInterceptorWithDependencies.childPostConstruct = false;
		// key setting
		DefaultChildInterceptorWithDependencies.shortCircuitPostConstruct = true;

		DeploymentUnitMetadata unit = new DeploymentUnitMetadata();
		Class clazz = DefaultNephewInterceptorWithDependencies.class;
		Method aroundInvokeMethod = JeeAnnotationUtils.findAroundInvokeMethod(clazz);
		InterceptionProcessor processor = new InterceptionProcessor();
		
		InterceptorMetadata defaultInterceptor = processor.createInterceptorMetadata(clazz, aroundInvokeMethod);
		unit.registerDefaultInterceptorMetadata(defaultInterceptor);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, tcc, unit);

		BusinessInterface bi = (BusinessInterface) jarAc.getBean(ChildSelfInterceptBean.class.getSimpleName());
		assertTrue("Should be proxied", AopUtils.isAopProxy(bi));

		// check super classes
		assertTrue("PreConstruct was not called", DefaultChildInterceptorWithDependencies.childPostConstruct);
		assertFalse("PreDestroy was called", DefaultChildInterceptorWithDependencies.childPreDestroy);

		// check base interceptor
		assertFalse("PreConstruct was called - shortcircuit was not considered", DefaultNephewInterceptorWithDependencies.nephewPostConstruct);
		assertFalse("PreDestroy was called", DefaultNephewInterceptorWithDependencies.nephewPreDestroy);

		// Programmatically raise event
		tcc.getDeployedComponents().get(0).invokeLifecycleMethods(bi, LifecycleEvent.PRE_DESTROY);

		// check base interceptor
		assertTrue("PreDestroy was not called", DefaultNephewInterceptorWithDependencies.nephewPreDestroy);

		// test super interceptor classes
		assertTrue("PreDestroy was not called", DefaultChildInterceptorWithDependencies.childPreDestroy);
	}
	
}
