/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.intercept;

import java.lang.reflect.Method;

import javax.interceptor.InvocationContext;

import junit.framework.TestCase;

import org.springframework.jee.inject.DeploymentUnitMetadata;

import sample.CountInterceptor;
import sample.EngineControlBean;
import sample.InterceptorWithDependencies;

/**
 * Unit tests for ordering on InterceptionMetadata
 * @author Rod Johnson
 *
 */
public class InterceptionMetadataTests extends TestCase {
	
	private InterceptionMetadata im;
	private DeploymentUnitMetadata dum;
	
	private Method interceptorWithDependenciesAroundInvokeMethod;
	private InterceptorMetadata interceptorWithDependencies;
	
	private Method countInterceptorAroundInvokeMethod;
	private InterceptorMetadata countInterceptor;
	
	protected DeploymentUnitMetadata createDeploymentUnitMetadata() {
		//GenericApplicationContext deploymentContext = new GenericApplicationContext();
		return new DeploymentUnitMetadata();
	}
	
	@Override
	protected void setUp() throws Exception {	
		dum = createDeploymentUnitMetadata();
		im = new InterceptionMetadata(dum, "ec", EngineControlBean.class);
		
		interceptorWithDependenciesAroundInvokeMethod = InterceptorWithDependencies.class.getMethod("countCall", new Class[] { InvocationContext.class });
		interceptorWithDependencies = new InterceptorMetadata(InterceptorWithDependencies.class, interceptorWithDependenciesAroundInvokeMethod);
		
		countInterceptorAroundInvokeMethod = CountInterceptor.class.getMethod("countCall", new Class[] { InvocationContext.class });
		countInterceptor = new InterceptorMetadata(CountInterceptor.class, countInterceptorAroundInvokeMethod);
	}
	
	public void testOrderingOfClassAndMethodInterceptors() throws SecurityException, NoSuchMethodException {
		assertTrue("No interceptors yet", im.getInterceptorMetadata().isEmpty());
		InterceptorMetadata countInterceptor = new InterceptorMetadata(InterceptorWithDependencies.class, 
				interceptorWithDependenciesAroundInvokeMethod, interceptorWithDependenciesAroundInvokeMethod);
		// Register method interceptor first
		im.registerInterceptorMetadata(countInterceptor);
		im.registerInterceptorMetadata(interceptorWithDependencies);
		
		assertEquals("2 interceptors now", 2, im.getInterceptorMetadata().size());
		
		assertSame("Class interceptor ordered first", interceptorWithDependencies, im.getInterceptorMetadata().get(0));
		assertSame("Method interceptor ordered last", countInterceptor, im.getInterceptorMetadata().get(1));
	}
	
	/**
	 * Check that class interceptors are processed in the order they
	 * are added with registerInterceptorMetadata method
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 */
	public void testOrderingAmongClassInterceptors() throws SecurityException, NoSuchMethodException {
		assertTrue("No interceptors yet", im.getInterceptorMetadata().isEmpty());
		InterceptorMetadata countInterceptor = new InterceptorMetadata(InterceptorWithDependencies.class, 
				interceptorWithDependenciesAroundInvokeMethod);
		assertEquals("Name defaults to FQN of the class", InterceptorWithDependencies.class.getName(), countInterceptor.getComponentName());
		
		// Register method interceptor first
		im.registerInterceptorMetadata(countInterceptor);
		im.registerInterceptorMetadata(interceptorWithDependencies);
		
		assertEquals("2 interceptors now", 2, im.getInterceptorMetadata().size());
		
		assertSame("First added class interceptor ordered first", countInterceptor, im.getInterceptorMetadata().get(0));
		assertSame("Second added class interceptor ordered last", interceptorWithDependencies, im.getInterceptorMetadata().get(1));
	}

}
