package org.springframework.jee.intercept;

import java.lang.reflect.Method;
import java.util.List;

import junit.framework.TestCase;

import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.jee.JeeAnnotationUtils;
import org.springframework.jee.inject.AnnotationComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.FieldInjection;
import org.springframework.jee.inject.Injection;
import org.springframework.jee.inject.Jsr250Processor;
import org.springframework.jee.server.Bootstrap;
import org.springframework.jee.transaction.BusinessInterface;
import org.springframework.jee.validation.InvalidDeploymentUnitException;

import sample.DefaultBaseInterceptorWithDependencies;
import sample.DefaultInterceptorWithDependencies;
import sample.EngineControl;
import sample.intercept.MultiInterceptorEngineControlBean;

/**
 * 
 * @author Costin Leau
 * 
 */
public class InterceptorInheritanceTests extends TestCase {

	private InterceptorMetadata baseInterceptorMetadata;

	protected void setUp() throws Exception {
		super.setUp();
		baseInterceptorMetadata = InterceptorInheritanceTests.createBaseInterceptorMetadata(DefaultNephewInterceptorWithDependencies.class);
	}

	protected void tearDown() throws Exception {
		super.tearDown();
		baseInterceptorMetadata = null;
	}

	public static InterceptorMetadata createBaseInterceptorMetadata(Class clazz) throws Exception {
		Method aroundInvokeMethod = JeeAnnotationUtils.findAroundInvokeMethod(clazz);
		if (aroundInvokeMethod == null)
			throw new IllegalArgumentException("no around invoke method was found");

		// create a dummy interceptorMetadata
		InterceptorMetadata interceptorMetadata = new InterceptorMetadata(clazz, aroundInvokeMethod);

		interceptorMetadata.setBeanDefinition(new RootBeanDefinition(clazz, true));

		Jsr250Processor processor = new Jsr250Processor();
		processor.process(interceptorMetadata);

		return interceptorMetadata;
	}

	public void testSuperInterceptorsInjection() throws Exception {
		List<Injection<?>> injections = baseInterceptorMetadata.getInjections();

		assertEquals("injections not determined correctly", 4, injections.size());

		Injection injection = injections.get(0);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("nephewEngineControl", injection.getName());

		injection = injections.get(1);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("childEngineControl", injection.getName());

		injection = injections.get(2);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("engineControl", injection.getName());

		injection = injections.get(3);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("engineControl", injection.getName());

	}

	public void testSuperInterceptorsInjectionTwoLevels() throws Exception {
		List<Injection<?>> injections = createBaseInterceptorMetadata(DefaultChildInterceptorWithDependencies.class).getInjections();

		assertEquals("injections not determined correctly", 3, injections.size());
		Injection injection = injections.get(0);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("childEngineControl", injection.getName());

		injection = injections.get(1);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("engineControl", injection.getName());

		injection = injections.get(2);
		assertSame(FieldInjection.class, injection.getClass());
		assertEquals("engineControl", injection.getName());
	}

	/**
	 * Test for verifying call hierarchy. Similar to the one from
	 * InterceptionTests.
	 * 
	 * @throws InvalidDeploymentUnitException
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 */
	public void testDefaultInterceptorWithHierarchy() throws InvalidDeploymentUnitException, SecurityException,
			NoSuchMethodException, NoSuchFieldException {

		DeploymentUnitMetadata deploymentUnitMetadata = new DeploymentUnitMetadata();

		deploymentUnitMetadata.registerDefaultInterceptorMetadata(baseInterceptorMetadata);

		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());

		DefaultInterceptorWithDependencies.count = 0;
		DefaultBaseInterceptorWithDependencies.countBase = 0;
		DefaultChildInterceptorWithDependencies.childCount = 0;
		DefaultNephewInterceptorWithDependencies.nephewCount = 0;

		ec.start();
		assertEquals("Business method is intercepted", 1, DefaultBaseInterceptorWithDependencies.countBase);
		assertEquals("Business method is intercepted", 1, DefaultInterceptorWithDependencies.count);
		assertEquals("Business method is intercepted", 1, DefaultChildInterceptorWithDependencies.childCount);
		assertEquals("Business method is intercepted", 1, DefaultNephewInterceptorWithDependencies.nephewCount);

	}

	public void testSuperInterceptorsTargetMethodCall() throws Exception {
		DeploymentUnitMetadata deploymentUnitMetadata = new DeploymentUnitMetadata();
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(baseInterceptorMetadata);

		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());

		MultiInterceptorEngineControlBean.startCalls = 0;
		ec.start();
		assertEquals("target method called multiple times", 1, MultiInterceptorEngineControlBean.startCalls);
	}

	public void testInterceptorsCallOrder() throws Exception {
		DeploymentUnitMetadata deploymentUnitMetadata = new DeploymentUnitMetadata();

		// clean up hardcoded injection
		InterceptorMetadata anotherInterceptor = createBaseInterceptorMetadata(DefaultChildInterceptorWithDependencies.class);

		// 1st is child interceptor
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(anotherInterceptor);

		// 2nd is nephew interceptor
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(baseInterceptorMetadata);

		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());

		List list = DefaultBaseInterceptorWithDependencies.order;

		list.clear();

		ec.start();

		assertEquals("call count", 7, list.size());
		// first interceptor/super interceptor chain
		assertSame(DefaultChildInterceptorWithDependencies.class, list.get(0));
		assertSame(DefaultChildInterceptorWithDependencies.class, list.get(1));
		assertSame(DefaultChildInterceptorWithDependencies.class, list.get(2));
		// second interceptor/super interceptor chain
		assertSame(DefaultNephewInterceptorWithDependencies.class, list.get(3));
		assertSame(DefaultNephewInterceptorWithDependencies.class, list.get(4));
		assertSame(DefaultNephewInterceptorWithDependencies.class, list.get(5));
		assertSame(DefaultNephewInterceptorWithDependencies.class, list.get(6));

	}

	public void testSuperInterceptorInstanceSharing() throws Exception {
		DeploymentUnitMetadata deploymentUnitMetadata = new DeploymentUnitMetadata();
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(baseInterceptorMetadata);

		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());

		DefaultChildInterceptorWithDependencies.anotherInstance = false;

		DefaultInterceptorWithDependencies.count = 0;
		DefaultBaseInterceptorWithDependencies.countBase = 0;
		DefaultChildInterceptorWithDependencies.childCount = 0;
		DefaultNephewInterceptorWithDependencies.nephewCount = 0;

		ec.start();

		assertFalse("instances not shared", DefaultChildInterceptorWithDependencies.anotherInstance);

		assertEquals("Business method is not intercepted", 1, DefaultBaseInterceptorWithDependencies.countBase);
		assertEquals("Business method is not intercepted", 1, DefaultInterceptorWithDependencies.count);
		assertEquals("Business method is not intercepted", 1, DefaultChildInterceptorWithDependencies.childCount);
		assertEquals("Business method is not intercepted", 1, DefaultNephewInterceptorWithDependencies.nephewCount);

	}

	public void testMultipleSuperInterceptorChains() throws Exception {
		DeploymentUnitMetadata deploymentUnitMetadata = new DeploymentUnitMetadata();
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(baseInterceptorMetadata);

		// clean up hardcoded injection
		InterceptorMetadata anotherInterceptor = createBaseInterceptorMetadata(DefaultChildInterceptorWithDependencies.class);

		// add superclass interceptor as normal default interceptor
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(anotherInterceptor);

		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());

		DefaultInterceptorWithDependencies.count = 0;
		DefaultBaseInterceptorWithDependencies.countBase = 0;
		DefaultChildInterceptorWithDependencies.childCount = 0;
		DefaultNephewInterceptorWithDependencies.nephewCount = 0;

		ec.start();

		assertEquals("Business method is not intercepted", 2, DefaultBaseInterceptorWithDependencies.countBase);
		assertEquals("Business method is not intercepted", 2, DefaultInterceptorWithDependencies.count);
		assertEquals("Business method is not intercepted", 2, DefaultChildInterceptorWithDependencies.childCount);
		assertEquals("Business method is not intercepted", 1, DefaultNephewInterceptorWithDependencies.nephewCount);

	}

	public void testSuperSelfIntercept() throws Exception {
		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(ChildSelfInterceptBean.class);

		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, new DeploymentUnitMetadata());
		BusinessInterface bean = (BusinessInterface) jarAc.getBean(ChildSelfInterceptBean.class.getSimpleName());

		BaseSelfInterceptBean.baseSelfCount = 0;
		SelfInterceptBean.selfCount = 0;
		ChildSelfInterceptBean.childSelfCount = 0;

		bean.simpleMethod();

		assertEquals("super self around invoke is not called ", 1, ChildSelfInterceptBean.childSelfCount);
		assertEquals("super self around invoke is not called ", 1, ChildSelfInterceptBean.baseSelfCount);
		assertEquals("super self around invoke is not called ", 1, ChildSelfInterceptBean.selfCount);
	}

}
