/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.server;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import javax.interceptor.InvocationContext;

import junit.framework.TestCase;

import org.springframework.aop.Advisor;
import org.springframework.aop.framework.Advised;
import org.springframework.aop.support.AopUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.jee.inject.AnnotationComponentContributor;
import org.springframework.jee.inject.ChainedProcessorsComponentContributor;
import org.springframework.jee.inject.ComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.FieldInjection;
import org.springframework.jee.inject.Jsr250Processor;
import org.springframework.jee.inject.LifecycleEvent;
import org.springframework.jee.inject.ResourceInfo;
import org.springframework.jee.intercept.InterceptorMetadata;
import org.springframework.jee.intercept.JeeInterceptorPointcutAdvisor;
import org.springframework.jee.intercept.__ProxyControl;
import org.springframework.jee.validation.InvalidDeploymentUnitException;

import sample.ContainerInterfaceImplementingEngineControlBean;
import sample.CountInterceptor;
import sample.DefaultBaseInterceptorWithDependencies;
import sample.DefaultInterceptorWithDependencies;
import sample.EngineControl;
import sample.EngineControlBean;
import sample.InterceptorWithDependencies;
import sample.InvalidInterceptedBeanWithNoBusinessInterfaces;
import sample.LaundryMachine;
import sample.SelfInterceptorEngineControlBean;
import sample.__DummyContainerControlInterface;
import sample.aspectj.TraceAspect;
import sample.intercept.InterceptedLaundryMachine;
import sample.intercept.MultiInterceptorEngineControlBean;
import sample.intercept.StartInterceptor;
import sample.intercept.StopInterceptor;
import ejbs.Echo;
import ejbs.SimpleEchoBean;

/**
 * 
 * @author Rod Johnson
 *
 */
public class InterceptionTests extends TestCase {
		
	protected Echo getSimpleEchoBean() throws InvalidDeploymentUnitException {
        Bootstrap bootstrap = new Bootstrap();
        ComponentContributor ms = new AnnotationComponentContributor(SimpleEchoBean.class, EngineControlBean.class);
        ApplicationContext jarAc = bootstrap.deploy(ms);
        return (Echo) jarAc.getBean(SimpleEchoBean.class.getSimpleName());
    }
	
	protected DeploymentUnitMetadata createDeploymentUnitMetadata() {
		//GenericApplicationContext deploymentContext = new GenericApplicationContext();
		return new DeploymentUnitMetadata();
	}
	
	public void testIsProxied() throws InvalidDeploymentUnitException {
		Echo e = getSimpleEchoBean();
		assertNotNull(e);
		assertTrue(AopUtils.isJdkDynamicProxy(e));
	}
	
	public void testInterceptorIsInjected() throws Throwable {
		InterceptorWithDependencies.count = 0;
		assertEquals(0, InterceptorWithDependencies.count);
		Echo e = getSimpleEchoBean();
		e.echo(null);
		assertEquals("Call must be intercepted", 1, InterceptorWithDependencies.count);
	}
	
	/**
	 * Test the case where the Interceptors annotation
	 * is on the component class itself
	 * @throws InvalidDeploymentUnitException 
	 */
	public void testSelfInterceptor() throws InvalidDeploymentUnitException {
        Bootstrap bootstrap = new Bootstrap();
        
        SelfInterceptorEngineControlBean.count = 0;

        ChainedProcessorsComponentContributor cc = new AnnotationComponentContributor(SelfInterceptorEngineControlBean.class);
        cc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
        ApplicationContext jarAc = bootstrap.deploy(cc);
        EngineControl sb = (EngineControl) jarAc.getBean(SelfInterceptorEngineControlBean.class.getSimpleName());
        //System.out.println(((Advised) sb).toProxyConfigString());
        
        assertEquals(0, SelfInterceptorEngineControlBean.count);
        sb.start();
        assertEquals("Self interceptor must apply", 1, SelfInterceptorEngineControlBean.count);
    }
	
	
	public void testInterceptorsOnMethodsAndClass() throws InvalidDeploymentUnitException {
		StartInterceptor.instantiated.clear();
	        
        Bootstrap bootstrap = new Bootstrap();
        ComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);
        ApplicationContext jarAc = bootstrap.deploy(cc);
        EngineControl sb = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());
        //System.out.println(((Advised) sb).toProxyConfigString());
        
        CountInterceptor.count = StartInterceptor.count = StopInterceptor.count = 0;
        assertEquals("Only one instance of StartInterceptor was instantiated eagerly", 1, StartInterceptor.instantiated.size());
        
        sb.start();
        assertEquals(1, CountInterceptor.count);
        assertEquals("Method-level interceptor was applied", 1, StartInterceptor.count);
        assertEquals(0, StopInterceptor.count);
        
        sb.stop();
        assertEquals(2, CountInterceptor.count);
        assertEquals(1, StartInterceptor.count);
        assertEquals(1, StopInterceptor.count);       
        
        // Check that the same interceptor instance is used again for StartInterceptor
        sb.startAgain();
        assertEquals(3, CountInterceptor.count);
        assertEquals("Same instance of method-level interceptor was applied again", 2, StartInterceptor.count);
        assertEquals("Stop interceptor not targeted", 1, StopInterceptor.count);
        
        sb.stop();
        assertEquals(4, CountInterceptor.count);
        assertEquals(2, StartInterceptor.count);
        assertEquals(2, StopInterceptor.count);     
        
        assertEquals("Only one instance of StartInterceptor was instantiated in total", 1, StartInterceptor.instantiated.size());
    }
	
	public void testDefaultBaseInterceptorWithDependencies() throws InvalidDeploymentUnitException, SecurityException, NoSuchMethodException, NoSuchFieldException {
		Bootstrap bootstrap = new Bootstrap();
		ComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);
		
		Method defaultBaseInterceptorWithDependenciesAroundInvokeMethod = DefaultBaseInterceptorWithDependencies.class.getMethod("countCallInheritance", new Class[] { InvocationContext.class });
		  
		InterceptorMetadata defaultBaseInterceptorWithDependencies = 
			new InterceptorMetadata(DefaultBaseInterceptorWithDependencies.class, defaultBaseInterceptorWithDependenciesAroundInvokeMethod);
		
		// Simulate analysis of injection annotations on interceptor class
		defaultBaseInterceptorWithDependencies.addInjection(new FieldInjection(
				DefaultBaseInterceptorWithDependencies.class.getDeclaredField("engineControl"), 
				new ResourceInfo("enxgineControl", null)));
		
		DeploymentUnitMetadata deploymentUnitMetadata = createDeploymentUnitMetadata();
		//deploymentUnitMetadata.associate(new GenericApplicationContext(), new GenericApplicationContext());
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(defaultBaseInterceptorWithDependencies);
		
		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());
		DefaultBaseInterceptorWithDependencies.countBase = 0;
		ec.start();
		assertEquals("Business method is intercepted", 1, DefaultBaseInterceptorWithDependencies.countBase);
	}

	/**
	 * Interceptors inheritance - more details in CoreContracts 12.3.1 (page 303).
	 * @throws InvalidDeploymentUnitException
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 */
	public void testDefaultInterceptorWithDependencies() throws InvalidDeploymentUnitException, SecurityException, NoSuchMethodException, NoSuchFieldException {
		Bootstrap bootstrap = new Bootstrap();
		AnnotationComponentContributor cc = new AnnotationComponentContributor(MultiInterceptorEngineControlBean.class);
		//AnnotationComponentContributor cc = new AnnotationComponentContributor(AnotherMultiInterceptorBean.class);
	    cc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
		
		Method defaultInterceptorWithDependenciesAroundInvokeMethod = DefaultInterceptorWithDependencies.class.getMethod("countCall", new Class[] { InvocationContext.class });
		  
		InterceptorMetadata defaultInterceptorWithDependencies = 
			new InterceptorMetadata(DefaultInterceptorWithDependencies.class, defaultInterceptorWithDependenciesAroundInvokeMethod);
		
		// Simulate analysis of injection annotations on interceptor class
		Jsr250Processor processor = new Jsr250Processor();
		processor.process(defaultInterceptorWithDependencies);
		
		DeploymentUnitMetadata deploymentUnitMetadata = createDeploymentUnitMetadata();
		//deploymentUnitMetadata.associate(new GenericApplicationContext(), new GenericApplicationContext());
		deploymentUnitMetadata.registerDefaultInterceptorMetadata(defaultInterceptorWithDependencies);
		
		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());
		DefaultInterceptorWithDependencies.count = 0;
		DefaultBaseInterceptorWithDependencies.countBase = 0;
		ec.start();
		assertEquals("Business method is intercepted", 1, DefaultBaseInterceptorWithDependencies.countBase);
		assertEquals("Business method is intercepted", 1, DefaultInterceptorWithDependencies.count);
	}


	public void testInterceptorsOfSameClassProduceSingleInstance() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
        TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(MultiInterceptorEngineControlBean.class);
        ApplicationContext jarAc = bootstrap.deploy(tcc);
        EngineControl sb = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());
		// Verify that we have the same instance for each JEE interceptor
        Map<Class, Object> interceptorsForBeanInstance1 = validateProxyChainLifecycle((Advised) sb);
		
		// Get another singleton, and check its interceptor chain is distinct
		EngineControl sb2 = (EngineControl) jarAc.getBean(MultiInterceptorEngineControlBean.class.getSimpleName());
		// Verify that we have the same instance for each JEE interceptor
		Map<Class, Object> interceptorsForBeanInstance2 = validateProxyChainLifecycle((Advised) sb2);
		
		// Assert that 
		for (Class interceptorClass : interceptorsForBeanInstance1.keySet()) {
			assertNotSame("Must have distinct instance of interceptor class " + interceptorClass.getName() + " for second proxy: " +
					"Had " + interceptorsForBeanInstance1.get(interceptorClass) + " and " + interceptorsForBeanInstance2.get(interceptorClass), 
					interceptorsForBeanInstance1.get(interceptorClass),
					interceptorsForBeanInstance2.get(interceptorClass));
		}
	}
	
	/**
	 * Check that there is only one instance of an interceptor with a given class
	 * @param proxy Spring AOP representing for intercepted object
	 * @return map from interceptor class to interceptor instance
	 */
	public static Map<Class, Object> validateProxyChainLifecycle(Advised proxy) {
		Map<Class, Object> interceptorInstances = new HashMap<Class, Object>();
		for (Advisor adv : proxy.getAdvisors()) {
			if (adv instanceof JeeInterceptorPointcutAdvisor) {
				JeeInterceptorPointcutAdvisor mipa = (JeeInterceptorPointcutAdvisor) adv;
				// Get instance, make sure it's the same
				Object instanceAlreadyFound = interceptorInstances.get(mipa.getAdvice().getJeeInterceptorInstance().getClass());
				if (instanceAlreadyFound == null) {
					interceptorInstances.put(mipa.getAdvice().getJeeInterceptorInstance().getClass(), mipa.getAdvice().getJeeInterceptorInstance());
				}
				else {
					assertSame(instanceAlreadyFound, mipa.getAdvice().getJeeInterceptorInstance());
				}
			}
		}
		return interceptorInstances;
	}
	
	public void testExplicitInvocationOfListenerMethodOnInterceptorClass() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ChainedProcessorsComponentContributor cc = new AnnotationComponentContributor(InterceptedLaundryMachine.class, EngineControlBean.class);
		cc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
        ApplicationContext jarAc = bootstrap.deploy(cc);
        LaundryMachine lm = (LaundryMachine) jarAc.getBean(InterceptedLaundryMachine.class.getSimpleName());

        assertNotNull("Was injected", lm.getEngineControl());
        
        assertFalse("Predestroy not yet called", lm.preDestroyCalled());
        
        __ProxyControl __pc = (__ProxyControl) lm;
        __pc.invokeLifecycleMethod(LifecycleEvent.PRE_DESTROY);
        assertTrue("Predestroy now called", lm.preDestroyCalled());
	}

	public void testAspectJAdviceAppliesToEjb() throws InvalidDeploymentUnitException {
        Bootstrap bootstrap = new Bootstrap();
        String[] pureSpringLocations = new String[] { "classpath:org/springframework/jee/server/aspectJTrace.xml" };
        ComponentContributor cc = new AnnotationComponentContributor(EngineControlBean.class);
        ApplicationContext jarAc = bootstrap.deploy(pureSpringLocations, null, cc, null);

        EngineControl engineControl = (EngineControl) jarAc.getBean(EngineControlBean.class.getSimpleName());
        TraceAspect traceAspect = (TraceAspect) jarAc.getBean(TraceAspect.class.getName());
         assertTrue(traceAspect.namesOfInvokedMethods.isEmpty());
        assertTrue(traceAspect.targets.isEmpty());
        engineControl.start();
        assertEquals(1, traceAspect.namesOfInvokedMethods.size());
        assertEquals(1, traceAspect.targets.size());
        assertTrue(traceAspect.namesOfInvokedMethods.contains("start"));
        assertFalse(traceAspect.namesOfInvokedMethods.contains("stop"));
        engineControl.stop();
        assertEquals(2, traceAspect.targets.size());
        assertTrue(traceAspect.targets.get(0) instanceof EngineControlBean);
        assertEquals(2, traceAspect.namesOfInvokedMethods.size());
        assertTrue(traceAspect.namesOfInvokedMethods.contains("start"));
        assertTrue(traceAspect.namesOfInvokedMethods.contains("stop"));
    }
	
	public void testContainerControlInterfaceCanBeRegisteredOnDeploymentUnitMetadata() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ComponentContributor cc = new AnnotationComponentContributor(ContainerInterfaceImplementingEngineControlBean.class);
        DeploymentUnitMetadata deploymentUnitMetadata = createDeploymentUnitMetadata();
        deploymentUnitMetadata.registerContainerControlInterface(__DummyContainerControlInterface.class);
        assertEquals(1, deploymentUnitMetadata.getContainerControlInterfaces().size());
        assertSame(__DummyContainerControlInterface.class, deploymentUnitMetadata.getContainerControlInterfaces().get(0));
        
		ApplicationContext jarAc = bootstrap.deploy(new String[0], null, cc, deploymentUnitMetadata);
		EngineControl ec = (EngineControl) jarAc.getBean(ContainerInterfaceImplementingEngineControlBean.class.getSimpleName());
		assertTrue("Is a proxy due to self interception", ec instanceof Advised);
		assertTrue("Proxy implements control interface", ec instanceof __DummyContainerControlInterface);
		__DummyContainerControlInterface cci = (__DummyContainerControlInterface) ec;
		ContainerInterfaceImplementingEngineControlBean.count = 0;
		ec.start();
		assertEquals("Business method is intercepted", 1, ContainerInterfaceImplementingEngineControlBean.count);
		assertTrue(cci.iAmATargetNotAnInterceptor());
		assertEquals("Method on control interface is not intercepted", 1, ContainerInterfaceImplementingEngineControlBean.count);
	}
	
	public void testDeploymentUnitIsAutomaticallyInvalidatedDueToNoBusinessInterfacesForInterceptedComponent() {		
        Bootstrap bootstrap = new Bootstrap();

        ComponentContributor cc = new AnnotationComponentContributor(InvalidInterceptedBeanWithNoBusinessInterfaces.class);
        try {
			bootstrap.deploy(cc);
			fail("Should have rejected invalid deployment of intercepted singleton without business interfaces");
		}
		catch (InvalidDeploymentUnitException ex) {
			// Ok
		}
	}
}
