/*
 * Copyright 2002-2006 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.jee.server;


import junit.framework.TestCase;

import org.springframework.aop.support.AopUtils;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.core.io.DefaultResourceLoader;
import org.springframework.core.io.ResourceLoader;
import org.springframework.jee.ejb.AtEjbLaundryMachineBean;
import org.springframework.jee.inject.AnnotationComponentContributor;
import org.springframework.jee.inject.ChainedProcessorsComponentContributor;
import org.springframework.jee.inject.ComponentContributor;
import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.jee.inject.LifecycleEvent;
import org.springframework.jee.validation.InvalidDeploymentUnitException;

import sample.ConstructorInjectedBean;
import sample.EngineControlBean;
import sample.IPostConstruct;
import sample.InjectedClassWithStaticFields;
import sample.LaundryMachine;
import sample.LaundryMachineBean;
import sample.Lookup;
import sample.LookupBean;
import sample.PostConstructBean;
import sample.PostConstructBeanRelyingOnInterceptor;
import sample.SpringInjected;
import sample.SpringInjectedBean;
import ejbs.Echo;
import ejbs.SimpleEchoBean;

/**
 * 
 * @author Rod Johnson
 *
 */
public class Jsr250InjectionTests extends TestCase {
	
	protected Echo getSimpleEchoBean() throws InvalidDeploymentUnitException {
        Bootstrap bootstrap = new Bootstrap();
        ComponentContributor ms = new AnnotationComponentContributor(SimpleEchoBean.class, EngineControlBean.class);
        ApplicationContext jarAc = bootstrap.deploy(ms);
        return (Echo) jarAc.getBean(SimpleEchoBean.class.getSimpleName());
    }
	
	protected DeploymentUnitMetadata createDeploymentUnitMetadata() {
		//GenericApplicationContext deploymentContext = new GenericApplicationContext();
		return new DeploymentUnitMetadata();
	}
	
	public void testStartup() throws InvalidDeploymentUnitException {
		Echo e = getSimpleEchoBean();
		assertNotNull(e);
	}
	
	public void testEjbPropertyInjectionWithAtResource() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ChainedProcessorsComponentContributor ms = new AnnotationComponentContributor(LaundryMachineBean.class, EngineControlBean.class);
		// change the default support (to remove the tx interception for example)
		ms.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
        ApplicationContext jarAc = bootstrap.deploy(ms);
        LaundryMachine lm = (LaundryMachine) jarAc.getBean(LaundryMachineBean.class.getSimpleName());
        assertFalse("Should not be proxied", AopUtils.isAopProxy(lm));
        assertNotNull("Was injected", lm.getEngineControl());
	}
	
	
	public void testUnsatisfiedPropertyInjectionWithAtResource() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		try {
			
			ComponentContributor ms = new AnnotationComponentContributor(LaundryMachineBean.class);
			ApplicationContext jarAc = bootstrap.deploy(ms);
			jarAc.getBean(LaundryMachineBean.class.getSimpleName());
			fail("Cannot inject");
		}
		// TODO might want to change this to something more specific
		catch (BeanCreationException ex) {
				// Ok
		}
	}
	
	/**
	 * Test for J2EE 5.2.3: ability to inject static members
	 * before invoking main.
	 * @throws InvalidDeploymentUnitException
	 */
	public void testInjectionOnMainClass() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ChainedProcessorsComponentContributor cc = new AnnotationComponentContributor(EngineControlBean.class);
        GenericApplicationContext jarAc = bootstrap.deploy(cc);
        // TODO inelegant
        DeploymentUnitMetadata dum = new DeploymentUnitMetadata();
        
        Jsr250Metadata jsr250Metadata = cc.findComponentMetadata("main", InjectedClassWithStaticFields.class, dum);
        jsr250Metadata.setComponentContext(jarAc);
        jsr250Metadata.inject(null);
        
        // Will fail unless the class was injected
        InjectedClassWithStaticFields.main(null);
        
        //assertNotNull("Static field was injected", InjectedClassWithStaticFields.)
	}
	
	public void testPostConstructOnComponentClass() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ComponentContributor ms = new AnnotationComponentContributor(PostConstructBean.class);
        ApplicationContext jarAc = bootstrap.deploy(ms);
        IPostConstruct postConstructBean = (IPostConstruct) jarAc.getBean(PostConstructBean.class.getSimpleName());
        assertTrue("Post construct method must be called", postConstructBean.postConstructCalled());
	}
	
	public void testPostConstructListenerMethodOnInterceptorClass() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		ComponentContributor cc = new AnnotationComponentContributor(PostConstructBeanRelyingOnInterceptor.class);
        ApplicationContext jarAc = bootstrap.deploy(cc);
        IPostConstruct postConstructBean = (IPostConstruct) jarAc.getBean(PostConstructBeanRelyingOnInterceptor.class.getSimpleName());
        assertTrue("Post construct method must be called on interceptor", postConstructBean.postConstructCalled());
	}
	
	public void testExplicitInvocationOfListenerMethodOnComponentClass() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(LaundryMachineBean.class, EngineControlBean.class);
		// change the default processor stack (and use only interception)
		tcc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
        ApplicationContext jarAc = bootstrap.deploy(tcc);
        LaundryMachine lm = (LaundryMachine) jarAc.getBean(LaundryMachineBean.class.getSimpleName());
        assertFalse("Should not be proxied", AopUtils.isAopProxy(lm));
        assertNotNull("Was injected", lm.getEngineControl());
        
        assertFalse("Predestroy not yet called", lm.preDestroyCalled());
        
        assertEquals("2 components must have been deployed", 2, tcc.getDeployedComponents().size());
        
        // Programmatically raise event
        tcc.getDeployedComponents().get(0).invokeLifecycleMethods(lm, LifecycleEvent.PRE_DESTROY);
        assertTrue("Predestroy now called", lm.preDestroyCalled());
	}
	
	public void testDeploymentUnitKnowsAboutDeployedComponents() throws InvalidDeploymentUnitException {
		Bootstrap bootstrap = new Bootstrap();
		TrackingAnnotationComponentContributor tcc = new TrackingAnnotationComponentContributor(LaundryMachineBean.class, EngineControlBean.class);
		// change the default processor stack (and use only interception)
		tcc.setMetadataProcessors(ChainedProcessorsComponentContributor.INTERCEPTION_PROCESSORS);
        ApplicationContext jarAc = bootstrap.deploy(tcc);
        LaundryMachine lm = (LaundryMachine) jarAc.getBean(LaundryMachineBean.class.getSimpleName());
        assertFalse("Should not be proxied", AopUtils.isAopProxy(lm));
        assertNotNull("Was injected", lm.getEngineControl());

        assertEquals("2 components must have been deployed", 2, tcc.getDeployedComponents().size());
        
        Jsr250Metadata md = tcc.getDeployedComponents().get(0);
        assertNotNull(md.getDeploymentUnitMetadata());
        assertEquals("Deployment unit knows about both deployed component", 2, md.getDeploymentUnitMetadata().getDeployedComponentMetadata().size());
	}
	
	public void testSpringPropertyInjectionOfEjbsAndProcessingOfExternalSpringFiles() throws InvalidDeploymentUnitException {
    	testSpringInjectionOfEjbsAndProcessingOfExternalSpringFiles("org/springframework/jee/server/spring-ejb-jar-list.xml", SpringInjectedBean.class);
    }
    
    public void testSpringConstructorInjectionOfEjbsAndProcessingOfExternalSpringFiles() throws InvalidDeploymentUnitException {
    	testSpringInjectionOfEjbsAndProcessingOfExternalSpringFiles("org/springframework/jee/server/spring-ejb-jar-list-constructor.xml", ConstructorInjectedBean.class);
    }

    private void testSpringInjectionOfEjbsAndProcessingOfExternalSpringFiles(final String springLocation, 
    		Class ejbClass) throws InvalidDeploymentUnitException {
        EjbBootstrap bootstrap = new EjbBootstrap();
        // Provide a Spring DD
        ResourceLoader rl = new DefaultResourceLoader() {
        	public org.springframework.core.io.Resource getResource(String location) {
        		if (location.equals(EjbBootstrap.SPRING_EJB_JAR_XML_LOCATION)) {
        			return super.getResource(springLocation);
        		}
        		else {
        			return super.getResource(location);
        		}
        	}
        };

        String[] pureSpringLocations = new String[] { "classpath:org/springframework/jee/server/springExternal.xml" };
        ComponentContributor cc = new AnnotationComponentContributor(ejbClass);
        ApplicationContext jarAc = bootstrap.deploy(pureSpringLocations, rl, cc, createDeploymentUnitMetadata());

        SpringInjected sb = (SpringInjected) jarAc.getBean(ejbClass.getSimpleName());
        assertEquals("Spring must have injected values", 3, sb.getStrings().size());
        for (int i = 0; i < sb.getStrings().size(); i++) {
        	assertEquals(i + "", sb.getStrings().get(i));
        }

        // Test injection of inline singleton
       assertEquals("springInjectedMe", sb.getTruePojoInline().getMessage());

       // Test injection of singleton defined in external Spring file
       assertEquals("springInjectedMe", sb.getTruePojoExternal().getMessage());
    }
    

    public void testSpringLookupMethodOnEjb() throws InvalidDeploymentUnitException {
        EjbBootstrap bootstrap = new EjbBootstrap();
        // Provide a Spring DD
        ResourceLoader rl = new DefaultResourceLoader() {
        	public org.springframework.core.io.Resource getResource(String location) {
        		if (location.equals(EjbBootstrap.SPRING_EJB_JAR_XML_LOCATION)) {
        			return super.getResource("org/springframework/jee/server/spring-ejb-jar-lookup.xml");
        		}
        		else {
        			return super.getResource(location);
        		}
        	}
        };

        Class ejbClass = LookupBean.class;
        String[] pureSpringLocations = new String[] { "classpath:org/springframework/jee/server/springExternal.xml" };
        ComponentContributor cc = new AnnotationComponentContributor(ejbClass);
        ApplicationContext jarAc = bootstrap.deploy(pureSpringLocations, rl, cc, null);

        Lookup sb = (Lookup) jarAc.getBean(ejbClass.getSimpleName());
        assertSame(jarAc.getBean("truePojoExternal"), sb.getTruePojo());
    }
    
}
