package org.springframework.jee.transaction;

import javax.ejb.ApplicationException;
import javax.ejb.CreateException;

import org.easymock.EasyMock;
import org.springframework.core.NestedRuntimeException;
import org.springframework.test.AbstractDependencyInjectionSpringContextTests;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.DefaultTransactionDefinition;
import org.springframework.transaction.support.SimpleTransactionStatus;

public class ApplicationExceptionManagerTests extends AbstractDependencyInjectionSpringContextTests {

	PlatformTransactionManager tmMock;

	// use this to create adhoc exceptions;
	private DelegateBean delegate;

	private BusinessInterface proxy;

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.framework.TestCase#setUp()
	 */
	@Override
	protected void onSetUp() throws Exception {
		tmMock = (PlatformTransactionManager) applicationContext.getBean("txManager");
		delegate = (DelegateBean) applicationContext.getBean("delegate");
		proxy = (BusinessInterface) applicationContext.getBean("statelessBean");
		EasyMock.reset(tmMock);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.framework.TestCase#tearDown()
	 */
	@Override
	protected void onTearDown() throws Exception {
		tmMock = null;
		delegate = null;
		proxy = null;
	}

	public void testSetupSanity() throws Exception {
		Object beanObject = applicationContext.getBean("statelessBean");
		assertSame(proxy, beanObject);
		assertTrue(beanObject instanceof BusinessInterface);
		assertFalse(beanObject instanceof DelegateStatelessBean);

	}

	public void testCheckedApplicationException() {
		// test 1st method
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_MANDATORY");
		TransactionStatus status = new SimpleTransactionStatus();
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// expect application exception
		tmMock.commit(status);

		EasyMock.replay(tmMock);

		final NoSuchMethodException markupException = new NoSuchMethodException();
		delegate.setDelegate(new DelegateBean() {

			@Override
			public void methodWithApplicationException() throws NoSuchMethodException {
				throw markupException;
				// throw new NoSuchMethodException();
			}
		});

		try {
			proxy.methodWithApplicationException();
			fail("expected exception");
		}
		catch (NoSuchMethodException e) {
			// expected
			assertSame(markupException, e);
		}

		EasyMock.verify(tmMock);
	}

	public void testUnexpectedSystemException() {
		// test 1st method
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_MANDATORY");
		TransactionStatus status = new SimpleTransactionStatus();
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// unexpected exception (considered system-ex)
		tmMock.rollback(status);

		EasyMock.replay(tmMock);

		final RuntimeException markupException = new NestedRuntimeException("gotcha!") {
		};
		delegate.setDelegate(new DelegateBean() {

			@Override
			public void methodWithApplicationException() throws NoSuchMethodException {
				throw markupException;
			}
		});

		try {
			proxy.methodWithApplicationException();
			fail("expected exception");
		}
		catch (Exception e) {
			// expected
			assertSame(markupException, e);
		}

		EasyMock.verify(tmMock);
	}

	public void testSystemExceptionEvenIfDeclared() {
		// test 2st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_NEVER");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// errors are system exception
		tmMock.rollback(status);

		EasyMock.replay(tmMock);

		final Error markupException = new Error();
		delegate.setDelegate(new DelegateBean() {

			@Override
			public void methodWithError() throws Error {
				throw markupException;
			}
		});

		try {
			proxy.methodWithError();
			fail("expected exception");
		}
		catch (Throwable e) {
			// expected
			assertSame(markupException, e);
		}
		EasyMock.verify(tmMock);
	}

	public void testApplicationExceptionTooGeneric() {
		// test 3st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_REQUIRES_NEW");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// unchecked exception are not considered inside the throws clause
		tmMock.rollback(status);

		EasyMock.replay(tmMock);

		final RuntimeException markupException = new NestedRuntimeException("gotcha!") {
		};
		delegate.setDelegate(new DelegateBean() {

			@Override
			public void methodWithException() throws Exception {
				throw markupException;
			}
		});

		try {
			proxy.methodWithException();
			fail("expected exception");
		}
		catch (Throwable e) {
			// expected
			assertSame(markupException, e);
		}
		EasyMock.verify(tmMock);
	}

	public void testMixedDeclaredApplicationAndSystemException() {
		// test 4st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_SUPPORTS");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// got checked app exception
		tmMock.commit(status);

		EasyMock.replay(tmMock);

		final InterruptedException markupException = new InterruptedException();
		delegate.setDelegate(new DelegateBean() {
			@Override
			public void methodWithMultipleExceptions() throws NoSuchFieldException, UnsupportedOperationException,
					IllegalArgumentException, InterruptedException, CreateException {
				throw markupException;
			}
		});

		try {
			proxy.methodWithMultipleExceptions();
		}
		catch (InterruptedException interrupt) {
			// expected
			assertSame(markupException, interrupt);
		}
		catch (NoSuchFieldException notExpected) {
			fail("exception not expected" + notExpected);
		}
		catch (CreateException createException) {
			fail("exception not expected" + createException);
		}
		EasyMock.verify(tmMock);
	}
	
	
	public void testMixedDeclaredApplicationAndSystemException2() {
		// test 4st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_SUPPORTS");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// got checked app exception
		tmMock.commit(status);

		EasyMock.replay(tmMock);

		final CreateException markupException = new CreateException();
		delegate.setDelegate(new DelegateBean() {
			@Override
			public void methodWithMultipleExceptions() throws NoSuchFieldException, UnsupportedOperationException,
					IllegalArgumentException, InterruptedException, CreateException {
				throw markupException;
			}
		});

		try {
			proxy.methodWithMultipleExceptions();
		}
		catch (InterruptedException interrupt) {
			fail("exception not expected" + interrupt);
		}
		catch (NoSuchFieldException notExpected) {
			fail("exception not expected" + notExpected);
		}
		catch (CreateException createException) {
			// expected
			assertSame(markupException, createException);

		}
		EasyMock.verify(tmMock);
	}

	public void testUndeclaredRuntimeException() {
		// test 5st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_REQUIRED");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// unchecked sys exception
		tmMock.rollback(status);

		EasyMock.replay(tmMock);

		final RuntimeException markupException = new NestedRuntimeException("gotcha!") {
		};
		delegate.setDelegate(new DelegateBean() {
			@Override
			public void simpleMethod() {
				throw markupException;
			}
		});

		try {
			proxy.simpleMethod();
		}
		catch (NestedRuntimeException e) {
			// expected
			assertSame(markupException, e);
		}
		EasyMock.verify(tmMock);
	}

	@ApplicationException(rollback = true)
	private static class UnexpectedApplicationRuntimeExceptionWithRollback extends RuntimeException {

	}

	public void testUndeclaredApplicationRuntimeExceptionWithRollback() {
		// test 5st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_REQUIRED");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// unchecked sys exception
		tmMock.rollback(status);

		EasyMock.replay(tmMock);

		final RuntimeException markupException = new UnexpectedApplicationRuntimeExceptionWithRollback();
		delegate.setDelegate(new DelegateBean() {
			@Override
			public void simpleMethod() {
				throw markupException;
			}
		});

		try {
			proxy.simpleMethod();
		}
		catch (RuntimeException e) {
			// expected
			assertSame(markupException, e);
		}
		EasyMock.verify(tmMock);
	}

	@ApplicationException()
	private static class UnexpectedApplicationRuntimeExceptionWithCommit extends RuntimeException {

	}

	public void testUndeclaredApplicationRuntimeExceptionWithCommit() {
		// test 5st method
		TransactionStatus status = new SimpleTransactionStatus();
		DefaultTransactionDefinition definition = new DefaultTransactionDefinition();
		definition.setPropagationBehaviorName("PROPAGATION_REQUIRED");
		EasyMock.expect(tmMock.getTransaction(definition)).andReturn(status);

		// unchecked sys exception
		tmMock.commit(status);

		EasyMock.replay(tmMock);

		final RuntimeException markupException = new UnexpectedApplicationRuntimeExceptionWithCommit();
		delegate.setDelegate(new DelegateBean() {
			@Override
			public void simpleMethod() {
				throw markupException;
			}
		});

		try {
			proxy.simpleMethod();
		}
		catch (RuntimeException e) {
			// expected
			assertSame(markupException, e);
		}
		EasyMock.verify(tmMock);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.springframework.test.AbstractDependencyInjectionSpringContextTests#getConfigLocations()
	 */
	@Override
	protected String[] getConfigLocations() {
		return new String[] { "/org/springframework/jee/transaction/tx-context.xml" };
	}
}
