package org.springframework.jee.transaction;

import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;

import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;

import junit.framework.TestCase;

import org.springframework.jee.inject.DeploymentUnitMetadata;
import org.springframework.jee.inject.Jsr250Metadata;
import org.springframework.jee.server.TrackingAnnotationComponentContributor;
import org.springframework.jee.transaction.TransactionalMetadata;

/**
 * Tests if the tx metadata is correctly parsed and added.
 * 
 * @author Costin Leau
 * 
 */
public class TransactionMetadataTests extends TestCase {

	private TrackingAnnotationComponentContributor contributor;

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.framework.TestCase#setUp()
	 */
	@Override
	protected void setUp() throws Exception {
		contributor = new TrackingAnnotationComponentContributor();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.framework.TestCase#tearDown()
	 */
	@Override
	protected void tearDown() throws Exception {
		contributor = null;
	}

	/**
	 * Test Stateless attributes.
	 * 
	 * @throws Exception
	 */
	public void testStatelessMethodAttributes() throws Exception {
		String componentName = "foo";
		Jsr250Metadata metadata = contributor.findComponentMetadata(componentName, DelegateStatelessBean.class,
				new DeploymentUnitMetadata());
		assertTrue("metadata not transactional", metadata instanceof TransactionalMetadata);
		TransactionalMetadata txMetadata = (TransactionalMetadata) metadata;

		Map<Method, TransactionAttribute> methodAttributes = txMetadata.getTransactionalMethodAttributes();
		
		assertEquals("does not consider only interface methods", 5, methodAttributes.size());
		Class[] noParam = new Class[] {};
		Method method = DelegateStatelessBean.class.getMethod("methodWithApplicationException", noParam);
		TransactionAttribute attribute = methodAttributes.get(method);
		assertNotNull(attribute);
		assertSame(TransactionAttributeType.MANDATORY, attribute.value());
		
		method = DelegateStatelessBean.class.getMethod("methodWithError", noParam);
		attribute = methodAttributes.get(method);
		assertNotNull(attribute);
		assertSame(TransactionAttributeType.NEVER, attribute.value());
		
		method = DelegateStatelessBean.class.getMethod("methodWithException", noParam);
		attribute = methodAttributes.get(method);
		assertNotNull(attribute);
		assertSame(TransactionAttributeType.REQUIRES_NEW, attribute.value());
		
		method = DelegateStatelessBean.class.getMethod("methodWithMultipleExceptions", noParam);
		attribute = methodAttributes.get(method);
		assertNotNull(attribute);
		assertSame(TransactionAttributeType.SUPPORTS, attribute.value());
		
		method = DelegateStatelessBean.class.getMethod("simpleMethod", noParam);
		attribute = methodAttributes.get(method);
		assertNull(attribute);
	}

	
	/**
	 * Check against singleton managed session beans
	 * @throws Exception
	 */
	public void testBeanManagedAttributes() throws Exception {
		String componentName = "foo";
		Jsr250Metadata metadata = contributor.findComponentMetadata(componentName, BeanManagedSessionBean.class,
				new DeploymentUnitMetadata());
		assertTrue("metadata not transactional", metadata instanceof TransactionalMetadata);
		TransactionalMetadata txMetadata = (TransactionalMetadata) metadata;

		Map<Method, TransactionAttribute> methodAttributes = txMetadata.getTransactionalMethodAttributes();
		
		assertEquals("there should be no transactional method", 0, methodAttributes.size());
	}
	
	
	public void testStatelessApplicationExceptions() {
		String componentName = "foo";
		Jsr250Metadata metadata = contributor.findComponentMetadata(componentName, DelegateStatelessBean.class,
				new DeploymentUnitMetadata());
		assertTrue("metadata not transactional", metadata instanceof TransactionalMetadata);
		TransactionalMetadata txMetadata = (TransactionalMetadata) metadata;

		List<Class<? extends Exception>> appExceptions = txMetadata.getApplicationCheckedExceptions();
		
		// check exceptions
		assertEquals(5, appExceptions.size());
		assertTrue(appExceptions.contains(NoSuchMethodException.class));
		assertTrue(appExceptions.contains(Exception.class));
		assertTrue(appExceptions.contains(NoSuchFieldException.class));
		assertTrue(appExceptions.contains(InterruptedException.class));
	}

}
